#pragma once

#include <BWAPI.h>
#include <BWTA.h>
#include "IBehaviour.h"
#include "MyUnitGroup.h"

namespace scai {
	struct AttackBehaviour : public IBehaviour {
		std::string nm;

		MyUnitGroup targets;
		int next_update;
		bool recently_switched;

		BWAPI::Position position;
		AttackBehaviour(const MyUnitGroup& units, BWAPI::Position position): IBehaviour(units), position(position), recently_switched(true) {}

		virtual void onStart() {
			log("Starting attack behaviour");
			for (int i = 0; i != units.size(); ++i)
				units.at(i)->attack(position);
		}

		virtual void draw() {
			if (!units.is_selected()) return;
			if ((units.size() != 0) && (targets.size() != 0)) {
				BWAPI::Position start = units.at(0)->getPosition();
				for (size_t i = 0; i != targets.size(); ++i) {
					BWAPI::Position end = targets.at(i)->getPosition();
					BWAPI::Broodwar->drawLineMap(start.x(), start.y(), end.x(), end.y(), BWAPI::Colors::Green);
					start = end;
				}
			}
			if (units.size() != 0) {
				for (size_t i = 0; i != units.size(); ++i) {
					BWAPI::Position p = units.at(i)->getPosition();
					BWAPI::Broodwar->drawCircleMap(p.x(), p.y(), 25, BWAPI::Colors::Green);
				}
			}
		}

		virtual Response onFrame() {

			units.remove_dead();
			if (units.isEmpty()) {
				log("No more units in attack behaviour");
				failure();
				return Response(FAILURE);
			}



			next_update--;
			if (next_update <= 0) {
				targets = potential_targets();
				next_update = 50;
			}

			
			/* Select new targets if idling! */
			targets.remove_dead();
			if (targets.size() != 0) {
				recently_switched = false;
				for (int i = 0; i != units.size(); ++i) {
					if (units.at(i)->getOrderTarget() != targets.at(0)) {
						units.at(i)->attack(targets.at(0));
					}
				}
			} else if(!recently_switched){
				//Check if we have the possibility of another base to attack
				if(scai::Global::i()->get_scout_location_amount() > 1 && units.size() > 0 ) {
					log("ATTACK: Changing region");
					BWAPI::Position pos = scai::Global::i()->get_next_base();
					BWAPI::Position center1 = BWTA::getRegion(BWAPI::TilePosition(pos))->getCenter();
					BWAPI::Position center2 = BWTA::getRegion(BWAPI::TilePosition(units.at(0)->getPosition()))->getCenter();

					//Check for similarity, we dont want to attack the same region we are in
					if(!(center1.x() == center2.x() && center1.y() == center2.y())) {
						recently_switched = true;
						position = pos;
						for (int i = 0; i != units.size(); ++i)
							units.at(i)->attack(position);
					}
				}
			}

			for (int i = 0; i != units.size(); ++i) {
				if (!units.at(i)->isIdle()) return Response(EXECUTING);
				if (units.at(i)->getDistance(position) > 150) return Response(EXECUTING);
			}

			
			BWAPI::Broodwar->printf("END attack behaviour");
			success();
			return Response(SUCCESS);
		}

		virtual void onAddUnits(const MyUnitGroup& units) {
			this->units.remove_dead();
			for (int i = 0; i != units.size(); ++i)
				units.at(i)->attack(position);
		}

		virtual const char* name() {
			std::stringstream nam;
			nam << "Attack: [" << units.size() << ", " << targets.size() << "]";
			nm = nam.str();
			return nm.c_str();
		}

		/**
		 *  Find any hostile units within the vision of owned units
		 */
		MyUnitGroup potential_targets() {
			if (units.size() == 0) return MyUnitGroup();
			MyUnitGroup targets = scai::Global::i()->all().hostiles().nearby(units.at(0)->getPosition(), units.at(0)->getType().sightRange());
			return sort_targets(targets);
		}

		float val(const BWAPI::Unit* unit) {
			float v = unit->getType().mineralPrice() + 1.25 * unit->getType().gasPrice();

			//Distance penalty is the average of the distance from each unit to the target
			float distance_penalty = 0.0f;
			for(size_t i = 0 ; i < units.size() ; i++) {
				distance_penalty -= (float)(unit->getPosition().getApproxDistance(units.at(i)->getPosition()));
			}
			distance_penalty /= (float)(units.size());

			//To lessen the distance penalty.
			distance_penalty /= (float)(2.0f);
			v -= distance_penalty;


			return v;
		}

		struct Target {
			float value;
			BWAPI::Unit* unit;
			bool operator < (const Target& other) {
				return value < other.value;
			}
		};
		MyUnitGroup sort_targets(const MyUnitGroup& targets) {
			std::vector<Target> ts;
			for (size_t i = 0; i != targets.size(); ++i) {
				Target t;
				t.value = val(targets.at(i));
				t.unit = targets.at(i);
				ts.push_back(t);
			}

			MyUnitGroup sorted;
			std::sort(ts.begin(), ts.end());
			for (size_t i = 0; i != ts.size(); ++i)
				sorted.add(ts[i].unit);
			return sorted;
		}
	};

	struct HarvestBehaviour : public IBehaviour {
		MyUnitGroup minerals;
		HarvestBehaviour(const MyUnitGroup& units, const MyUnitGroup& minerals): IBehaviour(units), minerals(minerals), next_patch(0){}
		int next_patch;
		int frameStart;
		int count;

		virtual void onStart() {
			if(minerals.size() == 0) {
				log("Some error in harvest behaviour, mineral batch is empty");
				return;
			}
			log("harvest onstart");
			frameStart = BWAPI::Broodwar->getFrameCount();
			count = 42;
			for (int i = 0; i != units.size(); ++i) {
				log("harvest trying nextpatch");
				BWAPI::Unit* patch = minerals.at(next_patch);
				log("nextpatch finished");
				units.at(i)->gather(patch);
				next_patch = (next_patch + 1) % minerals.size();
			}
		}

		virtual void onAddUnits(const MyUnitGroup& units) {
			for (int i = 0; i != units.size(); ++i) {
				BWAPI::Unit* patch = minerals.at(next_patch);
				units.at(i)->gather(patch);
				next_patch = (next_patch + 1) % minerals.size();
			}
		}

		virtual Response onFrame() {
			if(minerals.size() == 0) {
				log("SUCCESS: Harvestbehaviour has no minerals, report finished");
				return Response(SUCCESS);
			}
			if(BWAPI::Broodwar->getFrameCount() < (frameStart + count)) {
				for (int i = 0; i != units.size(); ++i) {
					BWAPI::Unit* patch = minerals.at(next_patch);
					units.at(i)->gather(patch);
					next_patch = (next_patch + 1) % minerals.size();
				}
			}

			return Response(EXECUTING);
		}
	};

	/*
	 *  This behaviour moves a group of units to a specified position.
	 *  It succeeds when the entire group has stopped moving.
	 */
	struct MoveToBehaviour : public IBehaviour {
		BWAPI::Position position;
		MoveToBehaviour(MyUnitGroup units, BWAPI::Position position): IBehaviour(units), position(position) {}

		virtual void onStart() {
			units.remove_dead();
			for (int i = 0; i != units.size(); ++i) {
				units.at(i)->move(position);
			}
		}

		virtual void draw() {
			if (units.is_selected()) {
				for (size_t i = 0; i != units.size(); ++i) {
					BWAPI::Position p = units.at(i)->getPosition();
					BWAPI::Broodwar->drawLineMap(p.x(), p.y(), position.x(), position.y(), BWAPI::Colors::Green);
				}
			}
		}

		virtual Response onFrame() {
			if (!(position.isValid())) {
				failure();
				return Response(FAILURE);
			}
			bool done = true;
			units.remove_dead();
			if (units.isEmpty()) {
				failure();
				return Response(FAILURE);
			}
			for (int i = 0; i != units.size(); ++i) {
				if (!(units.at(i)->isMoving()))
					units.at(i)->move(position);
			}

			for (int i = 0; i != units.size(); ++i) {
				if (units.at(i)->getDistance(position) > 320)
					done = false;
			}
			if (done) {
				success();
				return Response(SUCCESS);
			}
			return Response(EXECUTING);
		}
	};

	struct AttackMoveToBehaviour : public IBehaviour {
		BWAPI::Position position;
		AttackMoveToBehaviour(MyUnitGroup units, BWAPI::Position position): IBehaviour(units), position(position) {}

		virtual void onStart() {
			log("Starting attackmove behaviour");
			units.remove_dead();
			for (int i = 0; i != units.size(); ++i) {
				units.at(i)->attack(position);
			}
		}

		virtual void draw() {
			if (units.is_selected()) {
				for (size_t i = 0; i != units.size(); ++i) {
					BWAPI::Position p = units.at(i)->getPosition();
					BWAPI::Broodwar->drawLineMap(p.x(), p.y(), position.x(), position.y(), BWAPI::Colors::Green);
				}
			}
		}

		virtual Response onFrame() {
			if (!(position.isValid())) {
				failure();
				return Response(FAILURE);
			}
			bool done = true;
			units.remove_dead();
			if (units.isEmpty()) {
				failure();
				return Response(FAILURE);
			}
			for (int i = 0; i != units.size(); ++i) {
				if ((units.at(i)->isIdle()))
					units.at(i)->attack(position);
			}

			for (int i = 0; i != units.size(); ++i) {
				if (units.at(i)->getDistance(position) > 320)
					done = false;
			}
			if (done) {
				log("Attackmove finished for size: %d", units.size());
				success();
				return Response(SUCCESS);
			}
			return Response(EXECUTING);
		}
	};


	/*
	 *  This behaviour instructs a production building to train a single unit.
	 *  Assumption:  units contains only the production building.
	 *  The behaviour succeeds when the building is no longer producing (for whatever reason).
	 */
	struct TrainBehaviour : public IBehaviour {
		BWAPI::UnitType type;
		std::string nm;
		TrainBehaviour(MyUnitGroup units, BWAPI::UnitType type): IBehaviour(units), type(type) {
			nm = "Train: ";
			nm += type.getName();
			if (units.size() != 1)
				throw std::exception("Tried to train with no production building.");
		}

		virtual const char* name() { return nm.c_str(); }
		virtual void onStart() {
			if (!units.at(0)->train(type))
				failure();
			scai::Global::i()->unCommitMinerals(type.mineralPrice());
			scai::Global::i()->unCommitGas(type.gasPrice());
		}

		virtual Response onFrame() {
			if (units.size() == 0) {
				failure();
				return Response(FAILURE);
			} else if (!units.at(0)->isTraining()) {
				success();
				return Response(SUCCESS);
			}
			return Response(EXECUTING);
		}
	};


	/*
	 *  Instructs a worker (not zerg drones) to construct a building at the specified location.
	 *  Assumption: units contains only the worker
	 *  Succeeds when building is completed.
	 */
	struct BuildBehaviour : public IBehaviour {
		BWAPI::Position position;
		BWAPI::UnitType type;
		std::string nm;

		virtual void draw() {
			if (units.is_selected()) {
				int w = type.tileWidth() * 32;
				int h = type.tileHeight() * 32;
				BWAPI::Broodwar->drawBoxMap(position.x(), position.y(), position.x() + w, position.y() + h, BWAPI::Colors::Green);
			}
		}

		virtual const char * name () {
			return nm.c_str();
		}

		int frameStart;
		bool issued;
		BWAPI::Unit* building;

		BuildBehaviour(MyUnitGroup units, BWAPI::Position position, BWAPI::UnitType type)
			: IBehaviour(units)
			, position(position)
			, type(type)
			, issued(false)
			, building(NULL)
		{
			nm = "Build: ";
			nm += type.getName();
			if (units.size() != 1)
				throw std::exception("Tried to build without a worker");
		}

		virtual void onStart() {
			units.remove_dead();
			if (units.size() == 0) {
				return failure();
			}
			frameStart = BWAPI::Broodwar->getFrameCount();
			issued = units.at(0)->build(BWAPI::TilePosition(position), type);
		}

		virtual Response onFrame() {
			if (!issued) {
				if (units.size() == 0) {
					failure();
					return Response(FAILURE);
				}

				frameStart = BWAPI::Broodwar->getFrameCount();
				issued = units.at(0)->build(BWAPI::TilePosition(position), type);
				if (!issued) {	
					scai::Global::i()->unReserve(BWAPI::TilePosition(position), type.tileWidth(), type.tileHeight());
					position = BWAPI::Position(scai::Global::i()->getBuildLocationNear(BWAPI::TilePosition(position), type));
					if (position == BWAPI::TilePositions::None) failure();
				}
			}

			if (building == NULL) {
				MyUnitGroup g = scai::Global::i()->all().mine().mostRecent(type, frameStart);
				if (!g.isEmpty()) {
					building = g.at(0);
					removeUnits(units[0]);
					scai::Global::i()->unCommitMinerals(type.mineralPrice());
					scai::Global::i()->unCommitGas(type.gasPrice());
				}
			}

			if (building != NULL && building->isCompleted()) {
				success();
				return Response(SUCCESS);
			}
			return Response(EXECUTING);
		}
		virtual void onSuccess() {
			if(building != NULL) {
				Message msg;
				msg.id = BuildingCompleted;
				msg.put("building", MyUnitGroup(building));
				messages.push_back(msg);
			}
		}


	};

	struct ScoutBehaviour : public IBehaviour {
		BWAPI::Position current;
		std::vector<BWAPI::Position> locations;
		int index;



		ScoutBehaviour(MyUnitGroup units, std::vector<BWAPI::Position> locations)
			: IBehaviour(units)
			, locations(locations)
			, index(0)
		{
			if (units.size() == 0)
				throw std::exception("Tried to scout without units");
		}
		virtual void onStart() {
			units.remove_dead();
			current = locations[index];
			for(size_t i = 0 ; i < units.size() ; i++) {
				if(!units.at(i)->move(current))
					BWAPI::Broodwar->printf("Could not issue move command!");
				else
					BWAPI::Broodwar->printf("Move to %d, %d", current.x(), current.y());
			}
		}

		virtual Response onFrame() {
			// TODO: Hotfix to repair
			for(size_t j = 0 ; j < units.size() ; j++) {
				units.at(j)->move(current);
			}
			for(size_t i = 0 ; i < units.size() ; i++) {
				if(units.at(i)->getDistance(current) < 160){
					MyUnitGroup nearby = scai::Global::i()->all().hostiles().buildings().nearby(units.at(i)->getPosition(), units.at(i)->getType().sightRange());
					MyUnitGroup nearby2 = scai::Global::i()->all().mine().buildings().nearby(units.at(i)->getPosition(), units.at(i)->getType().sightRange());
					
					//inform executive of a new base location
					if(nearby.size() > 0) {
					
						Message msg;
						msg.id = ScoutLocation;
						msg.put("location", current);
						messages.push_back(msg);
					
					}
					if(nearby.size() > 0 || nearby2.size() > 0) {
						int i = 0;
						for(std::vector<BWAPI::Position>::const_iterator it = locations.begin() ; it != locations.end();) {
							if(i == index) {
								it = locations.erase(it);
								if(locations.size() == 0) {
									return Response(SUCCESS);
								}
								index = index%locations.size();
								break;
							} else {
								i++;
								it++;
							}
						}
					} else {
						index = (index+1)%locations.size();
					}
					current = locations[index];

				}
			}
			units.remove_dead();
			if (units.isEmpty() || locations.size() == 0) return Response(SUCCESS);
			return Response(EXECUTING);


		}


	};

	struct ExpandBehaviour : public IBehaviour {
		BWAPI::TilePosition location;
		ExpandBehaviour(const MyUnitGroup& worker, const BWAPI::TilePosition location): IBehaviour(worker), location(location) {
			std::vector<IBehaviour*> c;
			c.push_back(new scai::MoveToBehaviour(worker, BWAPI::Position(location)));
			c.push_back(new scai::BuildBehaviour(worker, BWAPI::Position(location), BWAPI::UnitTypes::Protoss_Nexus));
			IBehaviour* seq = new scai::Sequence(worker, c);
			children.push_back(seq);
		}

		virtual void onStart() {
			BWAPI::Broodwar->printf("SPAWN: Expand (%d, %d)", location.x(), location.y());
			children[0]->onStart();
		}
		virtual Response onFrame() { return children[0]->doFrame(); }
		virtual void onEnd()   { children[0]->onEnd(); }
	};

	struct ReaverWrapper : public IBehaviour {
		ReaverWrapper(const MyUnitGroup& units, IBehaviour* attack): IBehaviour(units) {
			children.push_back(attack);
		}

		virtual void onStart() { children[0]->onStart(); }
		virtual void onSuccess() { children[0]->onSuccess(); }
		virtual void onFailure() { children[0]->onFailure(); }
		virtual void onEnd() { children[0]->onEnd(); }

		virtual Response onFrame() {
			units.remove_dead();
			for (size_t i = 0; i != units.size(); ++i) {
				if (units.at(i)->getType() == BWAPI::UnitTypes::Protoss_Reaver) {
					int scarabs = units.at(i)->getScarabCount();
					scarabs += units.at(i)->getTrainingQueue().size();
					if (scarabs >= 2) continue;
					if (scai::Global::i()->getAvailableMinerals() < 50) continue;
					units.at(i)->train(BWAPI::UnitTypes::Protoss_Scarab);
				}
			}

			if (children.size() == 0) {
				failure();
				return Response(FAILURE);
			}
			return children[0]->onFrame();
		}
	};

	struct ArchonBehaviour: public IBehaviour {
		int frames;

		ArchonBehaviour(const MyUnitGroup& units) : IBehaviour(units), frames(0) {
			if(units.size() < 2) {
				throw std::exception("tried to create archon without enough high templars");
			}
		}
		virtual void onStart() {
			units.at(0)->useTech(BWAPI::TechTypes::Archon_Warp, units.at(1));
		}
		virtual Response onFrame() {
			frames++;
			if(!units.at(0)->isMoving()) {
				return Response(SUCCESS);
			}
			if(frames > 50) {
				units.at(0)->useTech(BWAPI::TechTypes::Archon_Warp, units.at(1));
				frames = 0;
			}
			return Response(EXECUTING);
		}
	};

}