#pragma once

#include "Database.h"
#include "Case.h"
#include "CompositeCase.h"
#include "CaseMatcher.h"

#include <vector>
#include <utility>

#include <time.h>

namespace scai {
	class CBRModule {
		typedef std::pair<float, size_t> MatchPair;
		void improve_match(std::vector<MatchPair>& matches, float match, size_t which, size_t amount = 5) {
			MatchPair mp(match, which);
			if (matches.size() < amount) {
				matches.push_back(std::make_pair(match, which));
			} else {
				for (std::vector<MatchPair>::iterator it = matches.begin(); it != matches.end(); ++it) {
					if (it->first < match) {
						std::swap(*it, mp);
					}
				}
			}
		}
		Case select(const std::vector<MatchPair>& matches, const std::vector<MatchPair>& comp_matches, const CaseLibrary& lib, const CompositeCaseLibrary& libc, float& match) {
			MatchPair a = matches[util::random(matches.size())];
			MatchPair b = comp_matches[util::random(comp_matches.size())];

			if (a.first > b.first) {
				match = a.first;
				return lib[a.second];
			} else {
				match = b.first;
				CompositeCase c = libc[b.second];
				return c.cases[util::random(c.cases.size())];
			}
		}
	public:
		CBRModule(const std::string& dbindex): db(dbindex) { }

		//Currently selects the "best" case, IE, the one with the highest matching value
		Case get_case(const Gamestate& state, float& matchv) {
			clock_t start = clock();
			CaseLibrary& lib = db.lib;
			CompositeCaseLibrary& lib_comp = db.lib_comp;
			clock_t end = clock();

			scai::log("Time to copy = %d", end - start);

			start = clock();
			std::vector<MatchPair> matches;
			for(size_t i = 0 ; i < lib.size() ; i++) {
				float matchvalue = match.plannerCompare(lib[i], state, 0.0f);
				improve_match(matches, matchvalue, i);
			}
			
			std::vector<MatchPair> comp_matches;
			for(size_t i = 0 ; i < lib_comp.size() ; i++) {
				float matchvalue = match.plannerCompare(lib_comp[i].composite_case, state, 0.0f);
				improve_match(comp_matches, matchvalue, i);
			}
			end = clock();
			scai::log("Time to select = %d", end - start);

			if (comp_matches.size() > 0 && matches.size() > 0) {
				return select(matches, comp_matches, lib, lib_comp, matchv);
			} else if (matches.size() > 0) {
				MatchPair m = matches[util::random(matches.size())];
				matchv = m.first;
				return lib[m.second];
			} else if (comp_matches.size() > 0) {
				size_t which_case = util::random(comp_matches.size());
				matchv = comp_matches[which_case].first;
				size_t which_solution = comp_matches[which_case].second;
				CompositeCase c = lib_comp[which_solution];
				return c.cases[util::random(c.cases.size())];
			}
			throw std::exception("No matches found!");
		}



		int total(const CompositeCaseLibrary& lib) {
			int tot = 0;
			for(size_t i = 0 ; i < lib.size() ; i++) {
				tot += lib[i].cases.size();
			}
			return tot;
		}

		void create_composite_database(const std::string& out_db) {
			create_compcases();
			db.write_database(out_db);
		}

		void test_matcher() {
			/* Testing crazy */
			CaseLibrary& lib = db.lib;

			Case a = lib[0];
			float match_aa = match.caseCompare(a, a, 0.0);
			scai::log("similarity(a, a) = %f", match_aa);

			Case b = lib[1];
			float match_bb = match.caseCompare(b, b, 0.0);
			scai::log("similarity(b, b) = %f", match_bb);

			float match_ab = match.caseCompare(a, b, 0.0);
			scai::log("similarity(a, b) = %f", match_ab);
			float match_ba = match.caseCompare(b, a, 0.0);
			scai::log("similarity(b, a) = %f", match_ba);
		}

		void create_compcases() {
			CaseLibrary& lib = db.lib;
			CompositeCaseLibrary& lib_comp = db.lib_comp;

			float equality_level = 0.90f;
			
			for(CaseLibrary::const_iterator it = lib.begin() ; it != lib.end();) {
				scai::log("Processing case %d / %d", it - lib.begin(), lib.end() - lib.begin());
				//First checks if the current case fits in a composite case
				bool flag = false;
				//log("Trying to match with existing composite cases");
				size_t best;
				float matchvalue = 0.0f;
				for(size_t j = 0 ; j < lib_comp.size() ; j++) {
					//found a good match, take it
					float matchvalue2 = match.caseCompare(*it, lib_comp[j].composite_case, 0.0f);
					if(matchvalue2 > matchvalue) {
						matchvalue = matchvalue2;
						best = j;
					}

				}
				if(matchvalue > equality_level) {
					lib_comp[best].append_case(*it);
					scai::log(" joined in clustered case with match: %.3f", matchvalue);
					it = lib.erase(it);
					flag = true;
				}

				//Else it tries to create a new composite case from the library
				if(!flag) {
					for(CaseLibrary::const_iterator it2 = it+1 ; it2 != lib.end();) {
						float matchvalue = match.caseCompare(*it, *it2, 0.0f);

						//found a good match, take it
						if(matchvalue > equality_level) {
							CompositeCase comp_case(*it, *it2);
							lib_comp.push_back(comp_case);
							scai::log(" created composite case with match: %.3f", matchvalue);
							lib.erase(it2);
							it = lib.erase(it);
							flag = true;
							break;
						} else {
							it2++;
						}
					}
				}
				if(!flag) {
					scai::log(" left case alone with match: %.3f", matchvalue);
					it++;
				}
			}
		}

		void update_win_ratio() {
			CaseLibrary& lib = db.lib;
			CompositeCaseLibrary& lib_comp = db.lib_comp;
			std::vector<std::string> root = db.get_root();

			//Set all ratio_updated to false, so we will iterate over every element only once
			for(size_t i = 0 ; i < lib.size() ; i++) {
				lib[i].ratio_updated = false;
			}

			for(size_t i = 0 ; i < lib_comp.size() ; i++) {
				for(size_t j = 0 ; j < lib_comp[i].cases.size() ; j++) {
					lib_comp[i].cases[j].ratio_updated = false;
				}
			}

			for(size_t r = 0 ; r < root.size() ; r++) {
				std::string current = root[r];
				Case temp_parent;
				temp_parent.match_id = "ROOT";
				scai::Case& u = find_case(current, lib, lib_comp);
				update(temp_parent, u, lib, lib_comp);
			}

			//Update every composite case now that we have
			for(size_t i = 0 ; i < lib_comp.size() ; i++) {
				lib_comp[i].composite_case.matches.clear();
				lib_comp[i].composite_case.wins = 0;
				lib_comp[i].composite_case.loss = 0;
				for(size_t j = 0 ; j < lib_comp[i].cases.size() ; j++) {
					for(std::map<std::string, int>::const_iterator it = lib_comp[i].cases[j].matches.begin() ; it != lib_comp[i].cases[j].matches.end() ; it++) {
						if(lib_comp[i].composite_case.matches[it->first] == -1 || lib_comp[i].composite_case.matches[it->first] == 1){
							continue;
						} else {
							lib_comp[i].composite_case.matches[it->first] = it->second;
							if(it->second == 1) {
								lib_comp[i].composite_case.wins++;
							} else if(it->second == -1) {
								lib_comp[i].composite_case.loss++;
							}
						}
					}
				}
			}

			for(size_t i = 0 ; i < lib_comp.size() ; i++) {
				log("Here are wins for i = %d, %d %d", i, lib_comp[i].composite_case.wins, lib_comp[i].composite_case.loss);
			}
			/*
			db.set_lib(lib);
			db.set_lib_comp(lib_comp);
			*/
		}

		void update(Case &parent, Case &case1, CaseLibrary &lib, CompositeCaseLibrary &lib_comp) {

			if(case1.child == "NO CHILD") {
				return;
			} else if(!case1.ratio_updated) {
				case1.ratio_updated = true;
				Case& child = find_case(case1.child, lib, lib_comp);
				case1.matches.clear();
				case1.wins = 0;
				case1.loss = 0;
				//If the child is part of a composite case, we need to calculate the entire composite case before adding up
				if(part_composite(child.match_id, lib, lib_comp)) {
					CompositeCase& temp = find_comp_case(child.match_id, lib, lib_comp);
					for(size_t i = 0 ; i < temp.cases.size() ; i++) {
						//dont want to end up in a loop here
						if(!(temp.cases[i].match_id == case1.match_id) && !(temp.cases[i].match_id == parent.match_id) && !temp.cases[i].ratio_updated){
							update(case1, temp.cases[i], lib, lib_comp);
						} 
					}

					for(size_t i = 0 ; i < temp.cases.size() ; i++) {	
						if(!(temp.cases[i].match_id == case1.match_id) && !(temp.cases[i].match_id == parent.match_id)){
							for(std::map<std::string, int>::const_iterator it = temp.cases[i].matches.begin() ; it != temp.cases[i].matches.end() ; it++) {
								if(case1.matches[it->first] == -1 || case1.matches[it->first] == 1){
									continue;
								} else {
									case1.matches[it->first] = it->second;
									if(it->second == 1) {
										case1.wins++;
									} else if(it->second == -1) {
										case1.loss++;
									}
								}
							}
						}
					}
				} else {
					//Case is not part of a composite case, so simpler job:
					update(case1, child, lib, lib_comp);
					for(std::map<std::string, int>::const_iterator it = child.matches.begin() ; it != child.matches.end() ; it++) {
						case1.matches[it->first] = it->second;
						if(it->second == 1) {
							case1.wins++;
						} else if(it->second == -1) {
							case1.loss++;
						}
					} 
				}				
			}
		}

		Case& find_case(std::string name, CaseLibrary &lib, CompositeCaseLibrary &lib_comp) {
			if(name == "NO CHILD") {
				throw 500;
			}
			for(size_t i = 0 ; i < lib.size() ; i++) {
				if(lib[i].match_id == name) {
					return lib[i];
				}
			}
			for(size_t i = 0 ; i < lib_comp.size() ; i++) {
				for(size_t j = 0 ; j < lib_comp[i].cases.size() ; j++) {
					if(lib_comp[i].cases[j].match_id == name) {
						return lib_comp[i].cases[j];
					}
				}
			}
			throw 42;
		}

		CompositeCase& find_comp_case(std::string name, CaseLibrary &lib, CompositeCaseLibrary &lib_comp) {
			for(size_t i = 0 ; i < lib_comp.size() ; i++) {
				for(size_t j = 0 ; j < lib_comp[i].cases.size() ; j++) {
					if(lib_comp[i].cases[j].match_id == name) {
						return lib_comp[i];
					}
				}
			}

		}

		bool part_composite(std::string name, CaseLibrary &lib, CompositeCaseLibrary &lib_comp) {
			for(size_t i = 0 ; i < lib_comp.size() ; i++) {
				for(size_t j = 0 ; j < lib_comp[i].cases.size() ; j++) {
					if(lib_comp[i].cases[j].match_id == name) {
						return true;
					}
				}
			}
			return false;
		}

	private:
		Database db;
		CaseMatcher match;
	};
}