#pragma once

#include "log.h"
#include <BWAPI.h>
#include <BWTA.h>

#include "ILayer.h"
#include "PlanItem.h"
#include "UnitManager.h"
#include "HarvestManager.h"
#include "BuildingManager.h"
#include "PlanExecutor.h"



namespace scai {
	class Executive : public ILayer {
		bool recieved;
		Communication& comm;

		HarvestManager hm;
		UnitManager um;
		BuildingManager bm;
		PlanExecutor pe;


		bool harvesting;
		int hightemplar_count;

		bool canTrain(BWAPI::UnitType unit) {
			scai::Global* g = scai::Global::i();
			return ((g->getAvailableMinerals() >= unit.mineralPrice()) &&
					(g->getAvailableGas() >= unit.gasPrice()) &&
					(!g->all().producers(unit).not_training().isEmpty()));
		}

		bool canBuild(BWAPI::UnitType unit) {
			scai::Global* g = scai::Global::i();
			return ((g->getAvailableMinerals() >= unit.mineralPrice()) &&
					(g->getAvailableGas() >= unit.gasPrice()));
		}
	public:

		virtual void onUnitCreate(BWAPI::Unit* unit) {
			um.onUnitCreate(unit);
			bm.onUnitCreate(unit);
		}
		virtual void onStart() {
			hm.onStart();
		}

		virtual void onFrame() {
			scai::Global::i()->getResman().onFrame();

			handle_messages();
			//log ("um::onFrame()");
			um.onFrame();
			//log ("hm::onFrame()");
			hm.onFrame();
			//log ("bm::onFrame()");
			bm.onFrame();
			//log ("  ::onFrame() finished");

			pe.print_plan(5, 5);
			um.print_allocations(15, 220);

			handle_automove();

			/*TODO: Currently one item at the time, fix behaviourId for concurrency*/
			handle_planitems();

			BWAPI::Broodwar->drawTextScreen(450, 15, "%.2f, %.2f", scai::Global::i()->getResman().getMineralRate(), scai::Global::i()->getResman().getGasRate());
			
		}
		
		//Method to automatically move newly built units to the ramp
		void handle_automove(){
			if(um.has_newly_created()) {
				MyUnitGroup automove = um.get_newly_created();

				BWAPI::Position pos;
				BWTA::Region* reg;
				BWTA::Chokepoint* choke;
				reg = BWTA::getRegion(BWAPI::Broodwar->self()->getStartLocation());
				choke = *reg->getChokepoints().begin();

				pos = choke->getCenter();
				
				//TODO: This is a hotfix to make sure reavers can attack when defending ramp
				if(automove.at(0)->getType() == BWAPI::UnitTypes::Protoss_Reaver) {
					if (scai::Global::i()->getAvailableMinerals() > 50){
						automove.at(0)->train(BWAPI::UnitTypes::Protoss_Scarab);
						automove.at(0)->train(BWAPI::UnitTypes::Protoss_Scarab);
					}
				}
				bool flag = false;
				//TODO: This is a hotfix to create archons, and yes its very ugly
				if(automove.at(0)->getType().getID() == BWAPI::UnitTypes::Protoss_High_Templar.getID()) {
					hightemplar_count++;
					if(hightemplar_count > 1) {
						MyUnitGroup all = scai::Global::i()->all().mine().attackers();
						for(size_t i = 0 ; i < all.size() ; i++) {
							//Find a high templar that is not the same as the one we had in automove
							if(all.at(i)->getType().getID() == BWAPI::UnitTypes::Protoss_High_Templar.getID()) {
								log("found a high templar id is %d, our id is %d", all.at(i)->getID(), automove.at(0)->getID());
								if(all.at(i)->getID() != automove.at(0)->getID()) {
									
									hightemplar_count = 0;
									MyUnitGroup archonGroup;
									archonGroup.add(all.at(i));
									archonGroup.add(automove.at(0));

									Message release_unit;
									release_unit.id = ReleaseUnit;
									release_unit.put("unit", archonGroup);
									comm.sendDn(this, release_unit);

									Message archon;
									archon.id = NewBehaviour;
									archon.put("behaviourId", -1);
									archon.putString("behaviour", "archon");
									archon.put("unit", archonGroup);
									comm.sendDn(this, archon);
									flag = true;
								}
							}
						}
					}
				} 
				if(flag) return;

				Message release_unit;
				release_unit.id = ReleaseUnit;
				release_unit.put("unit", automove);
				comm.sendDn(this, release_unit);

				Message ramp;
				ramp.id = NewBehaviour;
				ramp.put("behaviourId", -1);
				ramp.putString("behaviour", "moveto");
				ramp.put("position", pos);
				ramp.put("unit", automove);
				comm.sendDn(this, ramp);


			}
		}

		void handle_messages() {
			while (comm.hasNext(this)) {
				Message msg = comm.getNext(this);

				if (msg.id == scai::NewPlan) {
					if (pe.has_next()) BWAPI::Broodwar->printf("WARNING: Executive recieved plan while one was still executing");
					try {
						pe.new_plan(msg.getCast<Plan>("plan"));
					} catch (boost::bad_any_cast e) {
						BWAPI::Broodwar->printf("ERROR: Executive recieved message 'NewPlan' without key 'plan'");
					}
				} else if (msg.id == scai::BehaviourStatus){
					int behaviourId = msg.getCast<int>("behaviourId");
					std::string status = msg.getString("behaviourStatus");
					
					um.deallocate_all(behaviourId);
					if (status == "Failed") {
						pe.failed_item(behaviourId);
					} else {
						pe.finished_item(behaviourId);
					}
				} else if (msg.id == scai::RemoveUnit) {
					um.add_units(msg.getCast<MyUnitGroup>("units"));
				} else if (msg.id == scai::BuildingCompleted) {

					MyUnitGroup building = msg.getCast<MyUnitGroup>("building");
					log ("Completed building %s", building.at(0)->getType().getName().c_str());
					BWAPI::Broodwar->printf("Completed building %s", building.at(0)->getType().getName().c_str());
					if(building.containsType(BWAPI::UnitTypes::Protoss_Assimilator)) {
						hm.add_gas_location(building.at(0)->getPosition(), building);
					} else if(building.containsType(BWAPI::UnitTypes::Protoss_Nexus)) {
						hm.add_resource_location(building.at(0)->getPosition());
					}
				} else if(msg.id == scai::ScoutLocation) {
					BWAPI::Position pos = msg.getCast<BWAPI::Position>("location");
					scai::Global::i()->add_scouted_location(pos);
				} else {
					BWAPI::Broodwar->printf("WARNING: Executive recieved unknown message with id: %s", msg.id);
				}
			}
		}

		void handle_planitems() {
			if (pe.has_next()) {
				log ("PE has next true");
				PlanItem& next = pe.get_next();
				
				switch (next.myType) {
				case scai::Build:
					order_build(BWAPI::UnitType(next.build.type_id), next.behaviourId);
					break;
				case scai::Train:
					order_train(BWAPI::UnitType(next.train.type_id), next.behaviourId);
					break;
				case scai::Scout:
					order_scout(scai::Global::i()->get_all_scout_locations(), next.behaviourId);
					pe.finished_item(next.behaviourId);
					break;
				case scai::Attack:
					order_attack(scai::Global::i()->get_enemy_location(), scai::Global::i()->all().attackers(), next.behaviourId);
					pe.finished_item(next.behaviourId);
					break;
				case scai::Expand:
					order_build(BWAPI::UnitTypes::Protoss_Nexus, next.behaviourId, BWAPI::Position(scai::Global::i()->find_expansion(BWAPI::Broodwar->self()->getStartLocation())));
					pe.finished_item(next.behaviourId);
					break;
				case scai::Upgrade:
					order_upgrade(next.upgrade.upgrade_id, next.behaviourId);
					break;
				default:
					BWAPI::Broodwar->printf("WARNING: Unimplemented behaviour");
					pe.finished_item(next.behaviourId);
				}
			}
			//plan is empty, tell planning layer and delete it!
			else if (pe.is_empty()) {
				//TODO: use a plan id! not necessary right now with dummy impl
				log("INFO: plan empty");
				Message planFinished;
				planFinished.id = PlanFinished;
				planFinished.putString("planFinished", "Plan id?");
				comm.sendUp(this, planFinished);
				log("INFO: requested new plan");
			}
		}

		

		Executive(Communication & comm): hm(comm), um(hm), comm(comm), recieved(false), harvesting(false), hightemplar_count(0) {
			hm.executive = this;
		}

		/*
		void order_expand(const BWAPI::TilePosition& location, int behaviourId) {
			scai::Global::i()->commitMinerals(400);

			MyUnitGroup unit = um.get_worker();
			um.allocate(unit, behaviourId);
			
			Message get_builder;
			get_builder.id = ReleaseUnit;
			get_builder.put("unit", unit);
			comm.sendDn(this, get_builder);

			Message build;
			build.id = NewBehaviour;
			build.put("behaviourId", behaviourId);
			build.putString("behaviour", "expand");
			build.put("worker", unit);
			build.put("location", location);
			comm.sendDn(this, build);
		}
		*/

		void order_upgrade(int upgrade_id, int behaviour_id) {
			BWAPI::UpgradeType upgrade(upgrade_id);
			scai::MyUnitGroup upgraders = bm.get_upgraders(upgrade);
			if (upgraders.isEmpty()) {
				pe.failed_item(behaviour_id);
			} else {
				if(upgraders.at(0)->upgrade(upgrade))
					pe.finished_item(behaviour_id);
				else
					pe.failed_item(behaviour_id);
			}
		}

		void order_build(BWAPI::UnitType type, int behaviourId, BWAPI::Position pos = BWAPI::Positions::Unknown) {
			scai::Global::i()->commitMinerals(type.mineralPrice());
			scai::Global::i()->commitGas(type.gasPrice());


			MyUnitGroup unit = um.get_worker();
			um.allocate(unit, behaviourId);
			
			Message get_builder;
			get_builder.id = ReleaseUnit;
			get_builder.put("unit", unit);
			comm.sendDn(this, get_builder);

			Message build;
			build.id = NewBehaviour;
			build.put("behaviourId", behaviourId);
			build.putString("behaviour", "build");
			build.put("type", type);
			build.put("worker", unit);
			build.put("location", pos);
			comm.sendDn(this, build);
		}

		void order_train(BWAPI::UnitType type, int behaviour_id) {
			scai::Global::i()->commitMinerals(type.mineralPrice());
			scai::Global::i()->commitGas(type.gasPrice());

			MyUnitGroup producer = bm.get_producer(type);

			Message build;
			build.id = NewBehaviour;
			build.put("behaviourId", behaviour_id);
			build.putString("behaviour", "train");
			build.put("unit", type);
			build.put("producer", producer);
			comm.sendDn(this, build);

			//BWAPI::Broodwar->printf("Executive ordered train with id = %d", behaviour_id);
		}

		void order_scout(std::vector<BWAPI::Position> positions, int behaviour_id) {
			MyUnitGroup scouter = um.get_worker();
			
			Message get_builder;
			get_builder.id = ReleaseUnit;
			get_builder.put("unit", scouter);
			comm.sendDn(this, get_builder);

			Message scout;
			scout.id = NewBehaviour;
			scout.put("behaviourId", behaviour_id);
			scout.putString("behaviour", "scoutBaseLocation");
			scout.put("position", positions);
			scout.put("unit", scouter);
			comm.sendDn(this, scout);

		}

		void order_attack(BWAPI::Position pos, MyUnitGroup attackers, int behaviour_id) {

			Message get_attacker;
			get_attacker.id = ReleaseUnit;
			get_attacker.put("unit", attackers);
			comm.sendDn(this, get_attacker);

			Message attack;
			attack.id = NewBehaviour;
			attack.put("behaviourId", behaviour_id);
			attack.putString("behaviour", "attackBase");
			attack.put("position", pos);
			attack.put("unit", attackers);
			comm.sendDn(this, attack);
		}
	};
}