#pragma once

#include <vector>
#include "MyUnitGroup.h"

namespace scai {

	typedef std::map<std::string, boost::any> AnyMap;

	struct IBehaviour;
	typedef std::vector<IBehaviour*> Children;

	enum BTStatus {
		SUCCESS,
		FAILURE,
		EXECUTING,
		UNKNOWN
	};

	struct Response {
		Response(): status(UNKNOWN) {}
		Response(BTStatus status): status(status) {}
		BTStatus status;
		AnyMap reason;
	};

	struct IBehaviour {

		int print_behaviours(int x, int y, int indent = 0) {

			bool selected = units.is_selected();

			int lines = 1;
			BWAPI::Broodwar->drawTextScreen(x + 5 * indent, y, "%s%s: %d", (selected ? "\x06" : ""), name(), behaviourId);
			for (int i = 0; i != children.size(); ++i) {
				lines += children[i]->print_behaviours(x, y + 8 * lines, indent + 1);
			}
			return lines;
		}

		IBehaviour(MyUnitGroup units, Children children=Children()): units(units), children(children), behaviourId(-1) {}
		virtual ~IBehaviour() {
			for (int i = 0; i != children.size(); ++i) {
				delete children[i];
			}
		}

		virtual void draw() {
			/* Hook for behaviours to draw debug-info */
		}

		virtual const char * name() {
			return typeid(*this).name();
		}

		/**
		*  Behaviours _must_ override this method to implement their special features.
		*/
		virtual Response onFrame() = 0;

		/**
		*  Behaviours may override these methods to implement neccesary features.
		*/
		virtual void onStart() {}
		virtual void onSuccess() {}
		virtual void onFailure() {}
		virtual void onEnd() {}
		virtual Response onChildSuccess(IBehaviour * child) { return Response(EXECUTING); }
		virtual Response onChildFailure(IBehaviour * child) { return Response(EXECUTING); }
		virtual void onAddUnits(const MyUnitGroup& units) {}
		virtual void onRemUnits(const MyUnitGroup& units) {}


		virtual void addUnits(const MyUnitGroup& units) {
			this->units.add(units);
			onAddUnits(units);
		}

		/**
		*  Remove a group of units from this behaviour and its children.
		*/
		virtual void releaseUnits(const MyUnitGroup& unit) {
			units.release(unit);
			onRemUnits(unit);

			for (int i = 0; i != children.size(); ++i) {
				children[i]->releaseUnits(unit);
				children[i]->onRemUnits(unit);
			}
		}

		

		virtual Response doFrame() {
			preFrame();
			Response r;
			try {
				r = onFrame();
			} catch (std::exception e) {
				std::string err("Error in ");
				err += " " + err + name() + e.what() + "\n";
				throw std::exception(err.c_str());
			}
			postFrame();
			return r;
		}

		virtual void preFrame() {
			units.remove_dead();
		}

		virtual void postFrame() {
			draw();
			for (int i = 0; i != children.size(); ++i) {
				for (int j = 0; j != children[i]->messages.size(); ++j) {
					log ("Copying msg with id %d from %s up to %s", children[i]->messages[j].id, children[i]->name(), this->name());
					messages.push_back(children[i]->messages[j]);
				}
				children[i]->messages.clear();
			}
		}

		Children children;
		MyUnitGroup units;
		int behaviourId;
		std::vector<Message> messages;

	protected:
		virtual void removeUnits(const MyUnitGroup& remove) {
			units.release(remove);

			Message msg;
			msg.id = scai::RemoveUnit;
			msg.put("units", remove);

			messages.push_back(msg);
		}

		virtual void failure() {
			onFailure();
			onEnd();
		}
		virtual void success() {
			onSuccess();
			onEnd();
		}
	};


	struct Sequence : public IBehaviour {

		Sequence(MyUnitGroup units, std::vector<IBehaviour*> children)
			: IBehaviour(units, children)
			, current(0)
		{}

		virtual Response onFrame() {
			Response r = children[current]->doFrame();
			if (r.status == SUCCESS){
				return onChildSuccess(children[current]);
			}
			else if (r.status == FAILURE){
				return onChildFailure(children[current]);
			}
			return Response(EXECUTING);
		}

		virtual Response onChildSuccess(IBehaviour * child) {
			current++;
			if (current < children.size()) {
				
				children[current]->onStart();
				return Response(EXECUTING);
			} else {
				success();
				return Response(SUCCESS);
			}
		}

		virtual Response onChildFailure(IBehaviour * child) {
			failure();
			return Response(FAILURE);
		}

		virtual void onEnd() {
			current = 0;
		}

	private:
		size_t current;
	};

	struct Selector : public IBehaviour {

		Selector(MyUnitGroup units, Children children)
			: IBehaviour(units, children)
			, current(0)
		{ }

		virtual Response onFrame() {
			Response r = children[current]->doFrame();
			if (r.status == SUCCESS) {
				return onChildSuccess(children[current]);
			} else if (r.status == FAILURE) {
				return onChildFailure(children[current]);
			}
			return Response(EXECUTING);
		}

		virtual Response onChildSuccess(IBehaviour * child) {
			success();
			return Response(SUCCESS);
		}

		virtual Response onChildFailure(IBehaviour * child) {
			current++;
			if (current < children.size()) {
				
				children[current]->onStart();
				return Response(EXECUTING);
			} else {
				failure();
				return Response(FAILURE);
			}
		}


		virtual void onEnd() {
			current = 0;
		}

	private:
		size_t current;
	};

}