#pragma once

#include <exception>
#include <vector>
#include <algorithm>
#include <BWAPI.h>
#include "log.h"

#include "Predicates.h"

namespace scai {

	struct UnitData {
		UnitData(): age(0), assigned(false) {}
		UnitData(int age): age(age), assigned(false) {}

		int age;
		bool assigned;
	};

	int getAge(BWAPI::Unit* unit);

	/**
	 *  This represents a group of units under control of a behaviour.
	 */
	struct MyUnitGroup {

		/* Returns true if any of the contained units are selected */
		bool is_selected() {
			std::set<BWAPI::Unit*> selection = BWAPI::Broodwar->getSelectedUnits();
			for (std::set<BWAPI::Unit*>::const_iterator it = selection.begin(); it != selection.end(); ++it) {
				if (contains(*it)) return true;
			}
			return false;
		}

		void release(const MyUnitGroup& bye) {
			std::vector<BWAPI::Unit*> remove;
			for (int i = 0; i != bye.size(); ++i)
				remove.push_back(bye.units[i]);
			for (int i = 0; i != remove.size(); ++i) {
				std::vector<BWAPI::Unit*>::iterator it = std::find(units.begin(), units.end(), remove.at(i));
				if (it != units.end()) units.erase(it);
			}
		}

		bool isEmpty() {
			return units.size() == 0;
		}

		MyUnitGroup nearby(const BWAPI::Position& pos, int radius) {
			std::vector<BWAPI::Unit*> other;
			for (int i = 0; i != units.size(); ++i)
				if (units[i]->getDistance(pos) < radius)
					other.push_back(units[i]);
			return MyUnitGroup(other);
		}

		MyUnitGroup unassigned() {
			std::vector<BWAPI::Unit*> other;
			for (int i = 0; i != units.size(); ++i) {
				if (((UnitData*)units[i]->getClientInfo())->assigned == false) other.push_back(units[i]);
			}
			return MyUnitGroup(other);
		}

		MyUnitGroup closest(const BWAPI::Position& pos) {
			int c = 0;
			int min_dist = units[c]->getDistance(pos);

			for (int i = 0; i != units.size(); ++i) {
				int dist = units[i]->getDistance(pos);
				if (dist < min_dist) {
					min_dist = dist;
					c = i;
				}
			}
			return MyUnitGroup(units[c]);
		}


		//return the units that are within a certain range of the position
		MyUnitGroup withinRange(const BWAPI::Position& pos, int range) {
			MyUnitGroup temp;

			for(size_t i = 0 ; i < units.size() ; i++) {
				int dist = units[i]->getDistance(pos);
				if(dist < range){
					temp.add(units[i]);
				}
			}

			return temp;

		}

		MyUnitGroup mostRecent(const BWAPI::UnitType& type, int noOlderThan=0) {
			for (int i = 0; i != units.size(); ++i) {
				int age = ((UnitData*) units[i]->getClientInfo())->age;
				if (units[i]->getType() == type && age >= noOlderThan)
					return MyUnitGroup(units[i]);
			}
			return MyUnitGroup();
		}

		MyUnitGroup geysers() {
			return by_type(BWAPI::UnitTypes::Resource_Vespene_Geyser);
		}

		MyUnitGroup minerals() {
			return by_type(BWAPI::UnitTypes::Resource_Mineral_Field);
		}

		MyUnitGroup by_type(const BWAPI::UnitType& type) const {
			std::vector<BWAPI::Unit*> other;
			for (int i = 0; i != units.size(); ++i)
				if (units[i]->getType() == type) other.push_back(units[i]);
			return MyUnitGroup(other);
		}

		MyUnitGroup producers(const BWAPI::UnitType& type) {
			BWAPI::UnitType producer = type.whatBuilds().first;

			return by_type(producer);
		}

		MyUnitGroup not_training() {
			std::vector<BWAPI::Unit*> other;
			for (int i = 0; i != units.size(); ++i)
				if (!units[i]->isTraining()) other.push_back(units[i]);
			return MyUnitGroup(other);
		}

		MyUnitGroup mine() {
			std::vector<BWAPI::Unit*> others;
			for (int i = 0; i != units.size(); ++i)
				if (scai::IsMine(units[i])) others.push_back(units[i]);
			return MyUnitGroup(others);
		}

		MyUnitGroup hostiles() {
			std::vector<BWAPI::Unit*> others;
			for (int i = 0; i != units.size(); ++i)
				if (scai::IsEnemys(units[i])) others.push_back(units[i]);
			return MyUnitGroup(others);
		}

		MyUnitGroup harvesters() {
			std::vector<BWAPI::Unit*> others;
			for (int i = 0; i != units.size(); ++i)
				if (scai::IsHarvester(units[i])) others.push_back(units[i]);
			return MyUnitGroup(others);
		}

		// Returns any buildings which are no longer under construction
		MyUnitGroup is_completed() {
			std::vector<BWAPI::Unit*> completed;
			for (int i = 0; i != units.size(); ++i)
				if (!units[i]->isBeingConstructed()) {
					completed.push_back(units[i]);
				}
			return MyUnitGroup(completed);
		}

		MyUnitGroup is_trained() {
			std::vector<BWAPI::Unit*> completed;
			for (int i = 0; i != units.size(); ++i)
				if (units[i]->isCompleted()) {
					completed.push_back(units[i]);
				}
			return MyUnitGroup(completed);
		}

		bool contains(BWAPI::Unit* unit) const {
			return std::find(units.begin(), units.end(), unit) != units.end();
		}

		MyUnitGroup upgraders(const BWAPI::UpgradeType& upgrade) const {
			BWAPI::UnitType ut = upgrade.whatUpgrades();
			return by_type(ut);
		}

		MyUnitGroup not_upgrading() const {
			std::vector<BWAPI::Unit*> ret;
			for (std::vector<BWAPI::Unit*>::const_iterator it = units.begin(); it != units.end(); ++it) {
				if (!(*it)->isUpgrading()) ret.push_back(*it);
			}
			return ret;
		}

		// Return any dead units
		// Note that this actually modifies the unitgroup!
		MyUnitGroup remove_dead() {
			std::vector<BWAPI::Unit*> dead;
			for (std::vector<BWAPI::Unit*>::iterator it = units.begin(); it != units.end(); ++it) {
				if ((*it)->getHitPoints() <= 0)
					dead.push_back(*it);
			}
			rem(dead);
			return MyUnitGroup(dead);
		}

		MyUnitGroup buildings() {
			std::vector<BWAPI::Unit*> o;
			for (size_t i = 0; i != size(); ++i) {
				if (units[i]->getType().isBuilding()) o.push_back(units[i]);
			}
			return MyUnitGroup(o);
		}

		bool containsType(const BWAPI::UnitType type) {
			for(size_t i = 0 ; i < units.size() ; i++) {
				if(units[i]->getType() == type){
					return true;
				}
			}
			return false;
		}

		MyUnitGroup attackers() {
			MyUnitGroup r(this->units);
			r.rem(harvesters());
			r.rem(buildings());
			r.remove_dead();
			return r;
		}

		void add(const MyUnitGroup& other) {
			for (int i = 0; i != other.size(); ++i)
				add(other.at(i));
		}

		void add(BWAPI::Unit* unit) {
			if (std::find(units.begin(), units.end(), unit) == units.end())
				units.push_back(unit);
		}

		void rem(const MyUnitGroup& other) {
			for (int i = 0; i != other.size(); ++i)
				rem(other.at(i));
		}

		void rem(BWAPI::Unit* unit) {
			std::vector<BWAPI::Unit*>::iterator r = std::find(units.begin(), units.end(), unit);
			if (r != units.end())
				units.erase(r);
		}

		BWAPI::Unit* at(const std::vector<BWAPI::Unit*>::size_type& index) const {
			if (index < 0 || index >= units.size()) {
				log("ERROR: Tried to access index at %d, size is %d", index, units.size());
				throw std::exception("Invalid index in MyUnitGroup::at()");
			}
			return units[index];
		}

		std::vector<BWAPI::Unit*>::size_type size() const {
			return units.size();
		}

		
		MyUnitGroup(): units() {}
		MyUnitGroup(BWAPI::Unit* unit) {
			units.push_back(unit);
		}
		MyUnitGroup(const MyUnitGroup& other): units(other.units) {}
		MyUnitGroup(const std::vector<BWAPI::Unit*>& units): units(units) {}

		MyUnitGroup operator [] (const std::vector<BWAPI::Unit*>::size_type& index) {
			return MyUnitGroup(units[index]);
		}

		void clear() {
			units.clear();
		}

	private:
		std::vector<BWAPI::Unit*> units;
	};
}