#pragma once

#include <BWAPI.h>
#include "PlanItem.h"
#include <vector>
#include <algorithm>
#include "Resource.h"

namespace scai {

	class PlanExecutor {
		
		typedef std::vector<PlanItem> Plan;
		Plan current;

		/** 
		 *  Check which planitems this item depends on, if any
		 *  n is this items index in the current plan.
		 *  return behaviour_id of _first_ planitem this depends on or -1 if no dependencies.
		 */
		int check_dependency(const PlanItem& item, int n) {
			
			switch (item.myType) {
			case Train:
				return train_dependency(BWAPI::UnitType(item.train.type_id), n);
				break;
			case Build:
				return build_dependency(BWAPI::UnitType(item.build.type_id), n);
				break;
			case Upgrade:
				return upgrade_dependency(BWAPI::UpgradeType(item.upgrade.upgrade_id), n);
				break;
			case Attack:
				return (n == 0) ? -1 : current[n-1].behaviourId;
				break;
			case Defend:
				return -1;
				break;
			case Scout:
				return (n == 0) ? -1 : current[n-1].behaviourId;
				break;
			case Expand:
				return (n == 0) ? -1 : current[n-1].behaviourId;
				break;
			}
			return -1;
		}

		int upgrade_dependency(BWAPI::UpgradeType upgrade, int n) {
			return -1;
		}

		/**
		 *  Remove any dependencies which are accounted for in the given plan.
		 */
		void filter_planned(std::vector<BWAPI::UnitType>& deps, const Plan::const_iterator& begin, const Plan::const_iterator& end) {
			/* For each planned item */
			for (Plan::const_iterator it = begin; it != end; ++it) {
				/* For each dependency */
				for (std::vector<BWAPI::UnitType>::iterator dep = deps.begin(); dep != deps.end(); ++dep) {
					/* If planned item fulfils dependency, remove dependency and continue */
					if (it->myType == Build && it->build.type_id == dep->getID()) {
						dep = deps.erase(dep);
						if (dep == deps.end()) break;
					}
				}
			}
		}

		/**
		 *  Remove any dependencies which have already been met.
		 */
		void filter_done(std::vector<BWAPI::UnitType>& deps) {
			scai::Global* g = scai::Global::i();
			for (std::vector<BWAPI::UnitType>::iterator dep = deps.begin(); dep != deps.end();) {
				if (g->all().by_type(*dep).size() != 0) {
					scai::log("Removing dependency %s [DONE]", dep->getName().c_str());
					dep = deps.erase(dep);
				} else {
					++dep;
				}
			}
		}

		int build_dependency(BWAPI::UnitType building, int n) {
			scai::Global* g = scai::Global::i();
			std::vector<BWAPI::UnitType> reqs = g->missing_requirements(building);
			bool rdy = g->have_all(reqs);
			if (rdy) {
				return -1;
			} else {
				/*
				 Buildings only have a single dependency, so break at the first match.
				*/
				for (int i = 0; i != n; ++i) {
					PlanItem item = current[i];
					if (item.myType == Build && BWAPI::UnitType(item.build.type_id) == reqs[0])
						return item.behaviourId;
				}
			}
			return -1;
		}

		int train_dependency(BWAPI::UnitType unit, int n) {
			scai::Global* g = scai::Global::i();
			// The dependencies are already met
			std::vector<BWAPI::UnitType> reqs = g->missing_requirements(unit);
			bool rdy = g->have_all(reqs);
			if (rdy) {
				return -1;
			} else {
				/*
				 Units may depend on several buildings, so verify all are present, and depend on _last_ behaviour.
				 */
				int id = -1;
				for (int i = 0; i != n; ++i) {
					const PlanItem& item = current[i];
					
					if (item.myType == Build) {
						std::vector<BWAPI::UnitType>::iterator it = std::find(reqs.begin(), reqs.end(), BWAPI::UnitType(item.build.type_id));
						if (it != reqs.end()) {
							reqs.erase(it);
							id = item.behaviourId;
						}

					}
				}
				if (!reqs.size()) return id;
			}
			// Invalid plan
			std::string err = "Cannot train: ";
			err = err + unit.getName();
			throw std::exception(err.c_str());
		}

		/* HACK: Don't expand two times in a row */
		bool just_expanded;

	public:
		PlanExecutor(): just_expanded(false) {}

		std::string planitem_name(const PlanItem& item) {
			switch (item.myType) {
			case Build: return std::string("Build ") + BWAPI::UnitType(item.build.type_id).getName();
			case Train: return std::string("Train ") + BWAPI::UnitType(item.train.type_id).getName();
			case Upgrade: return std::string("Upgrade ") + BWAPI::UpgradeType(item.upgrade.upgrade_id).getName();
			case Attack: return std::string("Attack");
			case Defend: return std::string("Defend");
			case Expand: return std::string("Expand");
			case Scout: return std::string("Scout");
			default: return std::string("UNKNOWN");
			}
		}

		void print_plan(int x, int y) {
			
			for (int i = 0; i != current.size(); ++i) {
				Resource sum = cost(current.begin(), current.begin() + i + 1);
				BWAPI::Broodwar->drawTextScreen(x, y + 8 * i, "%d: %s [%d] [%d] [%d, %d, %d]", i, planitem_name(current[i]).c_str(), current[i].dependency,
					current[i].frameEstimate - scai::Global::i()->getCurrentFrame(),
					sum.minerals, sum.gas, sum.supply);
			}
		}

		void log_deps(const std::vector<BWAPI::UnitType>& deps) {
			for (size_t i = 0; i != deps.size(); ++i)
				scai::log("-- %s", deps[i].getName().c_str());
		}

		void log_plan(const Plan::const_iterator& begin, const Plan::const_iterator& end) {
			for (Plan::const_iterator it = begin; it != end; ++it)
				scai::log("-- %s", planitem_name(*it).c_str());
		}

		void new_plan(Plan plan) {
			//TODO: This is a hotfix! Some some plans do not end in a significant action, bug in tracer, append an attack if there is none
			PlanItem last = plan.at(plan.size()-1);
			if(!(last.myType == scai::Attack || last.myType == scai::Expand || last.myType == scai::Scout)) {
				PlanItem pi;
				pi.myType = scai::Attack;
				pi.attack.location = scai::Main;
				plan.push_back(pi);
			}


			/* Replace any train items (where count > 1) by count train items with count = 1 */
			for (Plan::iterator it = plan.begin(); it != plan.end(); ++it) {
				if (it->myType == scai::Train && it->train.count > 1) {
					int n = it->train.count;
					PlanItem foo(*it); foo.train.count = 1;
					it = plan.erase(it);
					plan.insert(it, n, foo);

					// Recheck plan for safetys sake.
					it = plan.begin(); 
				}
			}
			
			scai::Global* g = scai::Global::i();
			for (int i = 0; i != plan.size(); ++i) {
				std::vector<BWAPI::UnitType> dependencies = find_dependencies(plan[i]);

				if (cost(plan[i]).gas > 0) {
					dependencies.insert(dependencies.begin(), BWAPI::UnitTypes::Protoss_Assimilator);
				}
				if (cost(current.begin(), current.end()).supply > scai::Global::i()->get_supplyavailable()) {
					PlanItem item;
					item.myType = scai::Build;
					item.build.type_id = BWAPI::UnitTypes::Protoss_Pylon;
					add_at(item, current.size());
				} else if (cost(current.begin(), current.end()).supply > (scai::Global::i()->get_supplyavailable() + 9)) {
					/* This filters out excessive pylons */
					if (is_build(plan[i], BWAPI::UnitTypes::Protoss_Pylon)) {
						BWAPI::Broodwar->printf("Skipping excessive pylon");
						continue;
					}
				}
				if (g->all().mine().harvesters().size() > 20 * g->all().mine().by_type(BWAPI::UnitTypes::Protoss_Nexus).size()) {
					if (is_train(plan[i], BWAPI::UnitTypes::Protoss_Probe)) {
						BWAPI::Broodwar->printf("Skipping excessive probe");
						continue;
					}
				}
				if (is_build(plan[i], BWAPI::UnitTypes::Protoss_Assimilator) && !(g->can_build_assimilator())) {
					BWAPI::Broodwar->printf("Skipping excessive assimilator");
					continue;
				}
				if (plan[i].myType == Expand) {
					if (just_expanded) {
						just_expanded = false;
						continue;
					} else {
						just_expanded = true;
					}
				}

				filter_done(dependencies);
				filter_planned(dependencies, current.begin(), current.end());
				
				add_dependencies(dependencies);
				add(plan[i]);
			}
			/* TODO: Filter excessive pylons */
		}

		bool is_train(const PlanItem& item, const BWAPI::UnitType& type_id) {
			return item.myType == Train && item.train.type_id == type_id;
		}

		bool is_build(const PlanItem& item, const BWAPI::UnitType& type_id) {
			return item.myType == Build && item.build.type_id == type_id;
		}

		std::vector<BWAPI::UnitType> find_dependencies(const PlanItem& item) {
			if (item.myType == Build) {
				return scai::Global::i()->all_dependencies(BWAPI::UnitType(item.build.type_id));
			} else if (item.myType == Train) {
				return scai::Global::i()->all_dependencies(BWAPI::UnitType(item.train.type_id));
			} else {
				return std::vector<BWAPI::UnitType>();
			}
		}

		void failed_item(int behaviour_id) {
			for (Plan::iterator it = current.begin(); it != current.end(); ) {
				if (it->dependency == behaviour_id)
					it = current.erase(it);
				else
					++it;
			}
		}

		void finished_item(int behaviour_id) {
			for (int i = 0; i != current.size(); ++i) {
				if (current[i].dependency == behaviour_id)
					current[i].dependency = -1;
			}
		}

		bool is_empty() {
			return current.size() == 0;
		}

		bool has_next() {
			if (is_empty()) return false;
			if (can_execute(current.front(), 0)) return true;
			if (current.front().myType == Build && current.front().dependency == -1) return false;

			for (size_t i = 1; i < current.size(); ++i) {
				if (current[i].dependency != -1) continue;
				if (can_skip(current[i], i) && can_execute(current[i], i)) {
					Plan::iterator it = std::find(current.begin(), current.end(), current[i]);
					PlanItem copy(*it);
					current.erase(it);
					current.insert(current.begin(), copy);
					BWAPI::Broodwar->printf("Skipping ahead %s", planitem_name(copy).c_str());
					return true;
				}
			}
			return false;

		}

		PlanItem get_next() {
			PlanItem item = current.front();
			erase(item);
			for(size_t i = 0 ; i < current.size() ; i++) {
				if ( item.behaviourId == current[i].dependency) {
					current[i].frameEstimate = scai::Global::i()->getCurrentFrame() + scai::Global::i()->estimate_time(item);
					break;
				}
			}
			return item;
		}

		int estimateDelay(const PlanItem& item) {
			if (item.myType == Train) {
				MyUnitGroup producers = scai::Global::i()->all().mine().producers(BWAPI::UnitType(item.train.type_id));
				if (producers.size() == 0) {
					int rem = item.frameEstimate - scai::Global::i()->getCurrentFrame();
					return (rem < 0) ? 0 : rem;
				}
				int min = producers.at(0)->getRemainingTrainTime();
				for(size_t i = 1 ; i < producers.size() ; i++) {
					int temp = producers.at(i)->getRemainingTrainTime();
					if(temp < min) {
						min = temp;
					}
				}
				return min;
			} else if (item.myType == Build) {
				int rem = item.frameEstimate - scai::Global::i()->getCurrentFrame();
				return (rem < 0) ? 0 : rem;
			} else {
				return 0;
			}
		}

		bool can_skip(const PlanItem& item, int n) {
			//Noticed some cases where attacks were skipped ahead, should never be skipped
			if(item.myType == scai::Attack) {
				return false;
			}
			scai::Global* g = scai::Global::i();
			int frame = g->getCurrentFrame();

			int min = g->getAvailableMinerals();
			int gas = g->getAvailableGas();
			int sup = BWAPI::Broodwar->self()->supplyTotal() - BWAPI::Broodwar->self()->supplyUsed();
			for (int i = 0; i < n; ++i) {
				PlanItem test = current[i];
				int time = scai::Global::i()->estimate_time(test) + estimateDelay(test);
				min += (int) g->getMineralRate() * time;
				gas += (int) g->getGasRate() * time;
				int cost_m, cost_g, cost_s; scai::Global::i()->estimate_cost(test, cost_m, cost_g, cost_s);
				int my_cost_m, my_cost_g, my_cost_s;
				scai::Global::i()->estimate_cost(item, my_cost_m, my_cost_g, my_cost_s);
				if (!((min - cost_m >= my_cost_m) &&
					  (gas - cost_g >= my_cost_g) &&
					  (sup - cost_s >= my_cost_s))) {
					return false;
				}
				min -= cost_m;
				gas -= cost_g;
				sup -= cost_s;
			}
			
			return true;
		}

	private:

		Resource cost(const PlanItem& item) {
			scai::Global* g = scai::Global::i();
			Resource res;
			g->estimate_cost(item, res.minerals, res.gas, res.supply);
			return res;
		}

		Resource cost(Plan::const_iterator first, Plan::const_iterator end) {
			scai::Global* g = scai::Global::i();
			Resource res;
			for (Plan::const_iterator it = first; it != end; ++it) {
				Resource cur;
				g->estimate_cost(*it, cur.minerals, cur.gas, cur.supply);
				res += cur;
			}
			return res;
		}

		void add_dependencies(const std::vector<BWAPI::UnitType>& deps) {
			for (std::vector<BWAPI::UnitType>::const_iterator it = deps.begin(); it != deps.end(); ++it) {
				BWAPI::Broodwar->printf("Adding a dependency: %s", it->getName().c_str());
				PlanItem item;
				item.myType = Build;
				item.build.type_id = it->getID();
				add(item);
			}
		}

		void add_at(const PlanItem& item, size_t i) {
			/* We have not implemented behaviours to merge two templars, so skip these! */
			if (item.myType == Train && BWAPI::UnitType(item.train.type_id) == BWAPI::UnitTypes::Protoss_Archon)
				return;
			if (item.myType == Train && BWAPI::UnitType(item.train.type_id) == BWAPI::UnitTypes::Protoss_Scarab)
				return;

			Plan::iterator inserted = current.insert(current.begin() + i, item);
			inserted->behaviourId = scai::Global::i()->getNextId();
			inserted->dependency = check_dependency(*inserted, i);
		}

		/**
		 *  Add an item to the back of the plan, and assigns an id to this item.
		 */
		void add(const PlanItem& item) {
			add_at(item, current.size());
		}

		void add_front(const PlanItem& item) {
			add_at(item, 0);
		}

		void erase(const PlanItem& item) {
			Plan::iterator it = std::find(current.begin(), current.end(), item);
			if (it != current.end()) {
				size_t i = it - current.begin();
				current.erase(it);
			}
		}

		bool canTrain(BWAPI::UnitType unit) {
			scai::Global* g = scai::Global::i();

			return ((g->getAvailableMinerals() >= unit.mineralPrice()) &&
					(g->getAvailableGas() >= unit.gasPrice()) &&
					(!g->all().producers(unit).is_completed().not_training().isEmpty()) &&
					(g->hasSupply(unit)));
		}

		bool canBuild(BWAPI::UnitType unit) {
			scai::Global* g = scai::Global::i();
			return ((g->getAvailableMinerals() >= unit.mineralPrice()) &&
					(g->getAvailableGas() >= unit.gasPrice()));
		}

		bool canUpgrade(const BWAPI::UpgradeType& upgrade) {
			scai::Global* g = scai::Global::i();
			return ((g->getAvailableMinerals() >= upgrade.mineralPrice()) &&
				    (g->getAvailableGas() >= upgrade.gasPrice()) &&
					(!g->all().upgraders(upgrade).is_completed().not_upgrading().isEmpty()));
		}

		bool can_execute(const PlanItem& item, int n) {
			if (item.dependency != -1) return false;
			switch (item.myType) {
			case Build: return canBuild(BWAPI::UnitType(item.build.type_id));
			case Train: return canTrain(BWAPI::UnitType(item.train.type_id));
			case Expand: return canBuild(BWAPI::UnitTypes::Protoss_Nexus);
			case Upgrade: return canUpgrade(BWAPI::UpgradeType(item.upgrade.upgrade_id));
			default: return true;
			}
		}
	};
}