#pragma once

#include <string>
#include <vector>
#include <ostream>
#include <time.h>

#include "TestRunner.h"

namespace scai {
	struct TestRunPrinter {
		TestRunPrinter(const std::string& testfile) {
			runner.load_test(testfile);
		}

		void print(int level, std::ostream& out) {
			out << "Printing results for '" << runner.testfile << "' with level " << level << "\n\n";
			if (level >= 1) print1(out);
			if (level >= 2) print2(out);
			if (level >= 3) print3(out);
		}

		void print_timings(std::ostream& out) {
			std::vector<float> timings = collect_timings();
			out << "Timings based on " << timings.size() << " data points:" << std::endl;
			out << " Average = " << average(timings) << std::endl;
			out << " Min = " << minimum(timings) << ", Max = " << maximum(timings);
			out << ", stdev = " << stdev(timings) << std::endl;
		}

		void print_matches(std::ostream& out) {
			std::vector<float> matches = collect_matches();
			out << "Matches based on " << matches.size() << " data points:" << std::endl;
			out << " Average = " << average(matches) << std::endl;
			out << " Min = " << minimum(matches) << ", Max = " << maximum(matches);
			out << ", stdev = " << stdev(matches) << std::endl;
		}

	private:
		template <typename T>
		T average(const std::vector<T>& input) {
			T sum = 0;
			for (std::vector<T>::const_iterator it = input.begin(); it != input.end(); ++it)
				sum += *it;
			return sum / input.size();
		}

		template <typename T>
		T minimum(const std::vector<T>& input) {
			T min = input[0];
			for (std::vector<T>::const_iterator it = input.begin(); it != input.end(); ++it)
				min = std::min(min, *it);
			return min;
		}

		template <typename T>
		T maximum(const std::vector<T>& input) {
			T min = input[0];
			for (std::vector<T>::const_iterator it = input.begin(); it != input.end(); ++it)
				min = std::max(min, *it);
			return min;
		}

		template <typename T>
		T stdev(const std::vector<T>& input) {
			T avg = average(input);
			T accumulator = 0.0;
			for (std::vector<T>::const_iterator it = input.begin(); it != input.end(); ++it)
				accumulator += ((*it - avg) * (*it - avg));
			return std::sqrt(accumulator / input.size());
		}

		std::vector<float> collect_timings() {
			std::vector<float> timings;
			for (std::vector<TestRunner::TestData>::const_iterator it = runner.games.begin(); it != runner.games.end(); ++it) {
				for (std::vector<TestRunner::SelectionData>::const_iterator it2 = it->cases.begin(); it2 != it->cases.end(); ++it2) {
					timings.push_back(it2->time);
				}
			}
			return timings;
		}

		std::vector<float> collect_matches() {
			std::vector<float> matches;
			for (std::vector<TestRunner::TestData>::const_iterator it = runner.games.begin(); it != runner.games.end(); ++it) {
				for (std::vector<TestRunner::SelectionData>::const_iterator it2 = it->cases.begin(); it2 != it->cases.end(); ++it2) {
					matches.push_back(it2->match);
				}
			}
			return matches;
		}

		void print1(std::ostream& out) {
			
			out << "Runs " << count_all(runner) << "/" << runner.runs << "\n";
			if (count_all(runner) != runner.runs)
				out << "WARN: Testfile is missing " << (runner.runs - count_all(runner)) << " runs\n";
			
			int incomplete = count_where(runner, TestRunner::Unknown);
			if (incomplete != 0)
				out << "WARN: Testfile has " << incomplete << " incomplete runs\n\n";

			
			
			out << "Wins: " << count_wins(runner) << " (" << win_rate(runner) << "%)\n";
			out << "Losses: " << count_losses(runner) << " (" << loss_rate(runner) << "%)\n";

			out << "Win/Loss ratio: ";
			int losses = count_losses(runner);
			float wins = count_wins(runner);
			if (losses != 0)
				out << wins / (wins + losses);
			else
				out << "inf";
			out << '\n';

			int fails = count_where(runner, TestRunner::Crash);
			if (fails != 0) {
				out << "Crashes: " << fails << " (" << (100 * fails / count_all(runner)) << ")%\n";
			}
		}

		void print2(std::ostream& out) {
			std::cout << std::endl;
			for (std::vector<scai::TestRunner::TestData>::const_iterator it = runner.games.begin(); it != runner.games.end(); ++it) {
				std::cout << "Run (" << it->run << ") vs: " << it->race << " on " << it->map << " with outcome ";
				if (it->outcome == scai::TestRunner::Win) std::cout << "victory\n";
				else if (it->outcome == scai::TestRunner::Loss) std::cout << "loss\n";
				else if (it->outcome == scai::TestRunner::Crash) {
					std::cout << "crash\n";
					std::cout << "exception: " << it->error << "\n";
				}
				for (std::vector<scai::TestRunner::SelectionData>::const_iterator it2 = it->cases.begin(); it2 != it->cases.end(); ++it2) {
					std::cout << it2->time << ", ";
				}
				std::cout << std::endl;
			}
		}

		void print3(std::ostream& out) {

		}

		int count_where(const TestRunner& runner, int outcome) {
			int count = 0;
			for (size_t i = 0; i != runner.games.size(); ++i)
				if (runner.games[i].outcome == outcome)
					count += 1;
			return count;
		}

		int count_wins(const TestRunner& runner) {
			return count_where(runner, TestRunner::Win);
		}

		int win_rate(const TestRunner& runner) {
			int all = count_all(runner);
			if (all == 0) return 0;
			return 100 * count_wins(runner) / all;
		}

		int count_losses(const TestRunner& runner) {
			return count_where(runner, TestRunner::Loss);
		}

		int loss_rate(const TestRunner& runner) {
			int all = count_all(runner);
			if (all == 0) return 0;
			return 100 * count_losses(runner) / all;
		}

		int count_all(const TestRunner& runner) {
			return runner.games.size();
		}

		int count_successful(const TestRunner& runner) {
			return count_wins(runner) + count_losses(runner);
		}

		TestRunner runner;
	};
}