#pragma once

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/string.hpp>

namespace scai {
	enum UnitState {InTraining, Alive};
	enum BuildingState {UnderConstruction, Idle, Training, Researching};

	struct GameTrace {
		std::string map_name;
		size_t n_players;
		int frames;
		bool operator == (const GameTrace& o) const {
			return (
				(map_name == o.map_name) &&
				(n_players == o.n_players) &&
				(frames == o.frames)
			);
		}
		bool operator != (const GameTrace& o) const {
			return !(*this == o);
		}

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & map_name;
			ar & n_players;
			ar & frames;
		}
	};

	struct PlayerTrace {
		struct PlayerFrame {
			int frame;

			int minerals;
			int gas;
			int supply_used;
			int supply_total;

			bool operator == (const PlayerFrame& o) const {
				return ((minerals == o.minerals) && (gas == o.gas) &&
						(supply_used == o.supply_used) && (supply_total == o.supply_total));
			}

			bool operator != (const PlayerFrame& o) const {
				return !(*this == o);
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & frame;
				ar & minerals;
				ar & gas;
				ar & supply_used & supply_total;
			}
		};

		
		int id;
		int race_id;
		std::string race_name;
		std::string player_name;
		int start_x, start_y;

		std::vector<PlayerFrame> frames;

		bool operator == (const PlayerTrace& o) const {
			return (
				(id == o.id) &&
				(race_id == o.race_id) &&
				(start_x == o.start_x) && (start_y == o.start_y) &&
				(frames == o.frames)
			);
		}
		bool operator != (const PlayerTrace& o) const {
			return !(*this == o);
		}

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & id;
			ar & race_id;
			ar & race_name;
			ar & player_name;
			ar & start_x & start_y;
			ar & frames;
		}

		friend std::ostream& operator << (std::ostream& os, const PlayerTrace& t);
		friend std::istream& operator >> (std::istream& is, PlayerTrace& t);
	};

	struct UnitTrace {
		struct UnitFrame {
			int frame;
			int state;
			int health, shield;
			int pos_x, pos_y;
			int attacking;

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & frame;
				ar & state;
				ar & health & shield;
				ar & pos_x & pos_y;
				ar & attacking;
			}

			bool similar(const UnitFrame& o) {
				return ((state == o.state) && (health == o.health) && (shield == o.shield) &&
						(attacking == o.attacking) &&
						((std::abs(pos_x - o.pos_x) + std::abs(pos_y - o.pos_y)) < 25));
			}

			bool operator == (const UnitFrame& o) const {
				return ((state == o.state) && (health == o.health) && (shield == o.shield) &&
						(attacking == o.attacking) &&
						(pos_x == o.pos_x) && (pos_y == o.pos_y));
			}
			bool operator != (const UnitFrame& o) const {
				return !(*this == o);
			}
		};

		
		int id;
		int type_id;
		std::string type_name;
		int birth;
		int death;
		int owner;

		std::vector<UnitFrame> frames;
		UnitFrame get_at(int frame) {
			size_t cur = 0;
			while (cur < frames.size() - 1 && frames[cur].frame < frame) cur++;
			return frames[cur];
		}
		bool operator == (const UnitTrace& o) const {
			return (
				(id == o.id) &&
				(type_id == o.type_id) &&
				(birth == o.birth) &&
				(death == o.death) &&
				(owner == o.owner) &&
				(frames == o.frames)
			);
		}
		bool operator != (const UnitTrace& o) const {
			return !(*this == o);
		}

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & id;
			ar & type_id;
			ar & type_name;
			ar & birth & death;
			ar & owner;
			ar & frames;
		}

		friend std::ostream& operator << (std::ostream& os, const scai::UnitTrace& t);
		friend std::istream& operator >> (std::istream& is, scai::UnitTrace& t);
	};

	struct BuildingTrace {
		struct BuildingFrame {
			int frame;
			int state;
			bool operator == (const BuildingFrame& o) const {
				return state == o.state;
			}
			bool operator != (const BuildingFrame& o) const {
				return !(*this == o);
			}
			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & frame & state;
			}
		};

		int id;
		int type_id;
		std::string type_name;
		int owner;
		int birth, death;

		std::vector<BuildingFrame> frames;

		bool operator == (const BuildingTrace& o) const {
			return (
				(id == o.id) &&
				(type_id == o.type_id) &&
				(owner == o.owner) &&
				(birth == o.birth) &&
				(death == o.death) &&
				(frames == o.frames)
			);
		}
		bool operator != (const BuildingTrace& o) const {
			return !(*this == o);
		}

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & id;
			ar & type_id;
			ar & type_name;
			ar & owner;
			ar & birth & death;
			ar & frames;
		}

		friend std::ostream& operator << (std::ostream& os, const BuildingTrace& t);
		friend std::istream& operator >> (std::istream& is, BuildingTrace& t);
	};

	struct UpgradeTrace {
		int type_id;
		std::string type_name;
		int owner;
		int start, end;
		int level;

		bool operator == (const UpgradeTrace& o) const {
			return (
				(type_id == o.type_id) &&
				(owner == o.owner) &&
				(start == o.start) &&
				(end == o.end) &&
				(level == o.level)
			);
		}
		bool operator != (const UpgradeTrace& o) const {
			return !(*this == o);
		}

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & type_id;
			ar & type_name;
			ar & owner;
			ar & start & end;
			ar & level;
		}

		friend std::ostream& operator << (std::ostream& os, const UpgradeTrace& u);
		friend std::istream& operator >> (std::istream& is, UpgradeTrace& u);
	};

	struct Trace {

		GameTrace game;
		std::vector<PlayerTrace> players;
		std::vector<UnitTrace> units;
		std::vector<BuildingTrace> buildings;
		std::vector<UpgradeTrace> upgrades;
		std::string winner;

		bool operator == (const Trace& o) const {
			return (
				(game == o.game) &&
				(players == o.players) &&
				(units == o.units) &&
				(buildings == o.buildings) &&
				(upgrades == o.upgrades) &&
				(winner == o.winner)
			);
		}

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & game;
			ar & players;
			ar & units;
			ar & buildings;
			ar & upgrades;
			ar & winner;
		}
	};
}