#pragma once

#include <vector>
#include "MyUnitGroup.h"

namespace scai {

struct IBehaviour;
typedef std::vector<IBehaviour*> Children;


struct IBehaviour {

	IBehaviour(MyUnitGroup units, IBehaviour* parent, Children children=Children()): units(units), parent(parent), children(children), behaviourId(-1) {}
	virtual ~IBehaviour() {
		for (int i = 0; i != children.size(); ++i) {
			delete children[i];
		}
	}

	virtual const char * name() {
		return typeid(*this).name();
	}

	/**
	 *  Behaviours _must_ override this method to implement their special features.
 	 */
	virtual void onFrame() = 0;

	/**
	 *  Behaviours may override these methods to implement neccesary features.
	 */
	virtual void onStart() {}
	virtual void onSuccess() {}
	virtual void onFailure() {}
	virtual void onEnd() {}
	virtual void onChildSuccess(IBehaviour* child) {}
	virtual void onChildFailure(IBehaviour* child) {}

	/**
	 *  Remove a group of units from this behaviour and its children.
	 */
	virtual void releaseUnits(const MyUnitGroup& unit) {
		units.release(unit);
		for (int i = 0; i != children.size(); ++i) {
			children[i]->releaseUnits(unit);
		}
	}

	Children children;
	IBehaviour* parent;
	MyUnitGroup units;
	int behaviourId;

protected:
	void failure() {
		onFailure();
		onEnd();
		parent->onChildFailure(this);
	}
	void success() {
		onSuccess();
		onEnd();
		parent->onChildSuccess(this);
	}
};


struct Sequence : public IBehaviour {
	
	Sequence(MyUnitGroup units, IBehaviour* parent, std::vector<IBehaviour*> children)
		: IBehaviour(units, parent, children)
		, current(0)
	{
		for (int i = 0; i != children.size(); ++i)
			children[i]->parent = this;
	}

	virtual void onFrame() {
		children[current]->onFrame();
	}

	virtual void onChildSuccess(IBehaviour* child) {
		if (current < children.size()) {
			current++;
			children[current]->onStart();
		} else {
			success();
		}
	}

	virtual void onChildFailure(IBehaviour* child) {
		failure();
	}

	virtual void onEnd() {
		current = 0;
	}

private:
	int current;
};

struct Selector : public IBehaviour {
	
	Selector(MyUnitGroup units, IBehaviour* parent, Children children)
		: IBehaviour(units, parent, children)
		, current(0)
	{
		for (int i = 0; i != children.size(); ++i)
			children[i]->parent = this;
	}

	virtual void onFrame() {
		children[current]->onFrame();
	}

	virtual void onChildSuccess(IBehaviour* child) {
		success();
	}

	virtual void onChildFailure(IBehaviour* child) {
		if (current < children.size()) {
			current++;
			children[current]->onStart();
		} else {
			failure();
		}
	}

	virtual void onEnd() {
		current = 0;
	}

private:
	int current;
};

}