#pragma once

#include <algorithm>
#include <vector>
#include <map>
#include <set>
#include <fstream>
#include <BWAPI.h>
#include "log.h"
#include "Trace.h"

namespace scai {
	class Tracer {
	protected:
		std::string trace_name;

		// Maps players to indices in a trace
		std::map<BWAPI::Player*, size_t> players;

		// Which units are we currently tracing?
		std::vector<BWAPI::Unit *> traced_units;
		// Map from unit to index in trace
		std::map<BWAPI::Unit *, size_t> units;


		std::vector<BWAPI::Unit *> traced_buildings;
		std::map<BWAPI::Unit *, size_t> buildings;

		// Player_id, Upgrade_id
		std::vector<std::pair<int, int> > is_upgrading;
		
		void player_on_frame(int frame, BWAPI::Player* player, PlayerTrace& trace) {
			PlayerTrace::PlayerFrame f;
			f.frame = frame;
			f.minerals = player->minerals();
			f.gas = player->gas();
			f.supply_used = player->supplyUsed();
			f.supply_total = player->supplyTotal();

			if (trace.frames.size() == 0 || f != trace.frames.back()) {
				trace.frames.push_back(f);
			}
		}

		void create_player(BWAPI::Player* player) {
			if (player->isNeutral()) return;

			PlayerTrace p;
			p.id = player->getID();
			p.race_id = player->getRace().getID();
			p.race_name = player->getRace().getName();
			p.player_name = player->getName();
			p.start_x = player->getStartLocation().x();
			p.start_y = player->getStartLocation().y();

			players[player] = trace.players.size();
			trace.players.push_back(p);
		}

		void create_unit(BWAPI::Unit* unit, int frame) {
			if (unit->getPlayer() == NULL) return;
			if (unit->getPlayer()->isNeutral()) return;

			UnitTrace t;

			t.id = unit->getID();
			t.type_id = unit->getType().getID();
			t.type_name = unit->getType().getName();
			t.birth = frame;
			t.death = -1;
			t.owner = unit->getPlayer()->getID();

			units[unit] = trace.units.size();
			trace.units.push_back(t);
		}

		void unit_on_frame(BWAPI::Unit* unit, int frame, UnitTrace& t) {
			if (unit->getHitPoints() > 0) {
				UnitTrace::UnitFrame f;
				f.frame = frame;
				f.state = (unit->isBeingConstructed()) ? InTraining : Alive;
				f.health = unit->getHitPoints();
				f.shield = unit->getShields();
				f.pos_x = unit->getPosition().x();
				f.pos_y = unit->getPosition().y();
				BWAPI::Unit* target = unit->getOrderTarget(); // TODO: Use this or getTarget() ??
				f.attacking = (unit->isAttacking() && (target != NULL)) ? target->getID() : -1;

				if (t.frames.size() == 0 || !f.similar(t.frames.back())) {
					t.frames.push_back(f);
				}
			}
		}

		void create_building(BWAPI::Unit* unit, int frame) {
			if (unit->getPlayer() == NULL) return;
			if (unit->getPlayer()->isNeutral()) return;

			BuildingTrace t;

			t.id = unit->getID();
			t.type_id = unit->getType().getID();
			t.type_name = unit->getType().getName();
			t.owner = unit->getPlayer()->getID();
			t.birth = frame;
			t.death = -1;

			buildings[unit] = trace.buildings.size();
			trace.buildings.push_back(t);
		}

		void building_on_frame(BWAPI::Unit* unit, int frame, BuildingTrace& t) {
			BuildingTrace::BuildingFrame f;
			f.frame = frame;
			f.state = get_building_state(unit);
			if (t.frames.size() == 0 || f != t.frames.back())
				t.frames.push_back(f);
		}

		int get_building_state(BWAPI::Unit* unit) {
			if (unit->isBeingConstructed()) return UnderConstruction;
			else if (unit->getTrainingQueue().size() != 0) return Training;
			else if (unit->getRemainingUpgradeTime() != 0) return Researching;
			else return Idle;
		}

		void check_upgrades() {
			int frame = BWAPI::Broodwar->getFrameCount();
			std::set<BWAPI::Player *> players = BWAPI::Broodwar->getPlayers();
			std::set<BWAPI::UpgradeType> upgrades = BWAPI::UpgradeTypes::allUpgradeTypes();
			for (std::set<BWAPI::Player *>::iterator pl = players.begin(); pl != players.end(); ++pl) {
				for (std::set<BWAPI::UpgradeType>::iterator up = upgrades.begin(); up != upgrades.end(); ++up) {
					if ((*pl)->isUpgrading(*up)) {
						std::pair<int, int> plup = std::make_pair<int, int>((*pl)->getID(), up->getID());
						std::vector<std::pair<int, int> >::iterator it = std::find(is_upgrading.begin(), is_upgrading.end(), plup);
						if (it == is_upgrading.end()) {
							onUpgradeStart(*pl, *up, frame);
							is_upgrading.push_back(plup);
						}
					} else {
						std::vector<std::pair<int, int> >::iterator it = std::find(is_upgrading.begin(), is_upgrading.end(), std::make_pair<int, int>((*pl)->getID(), up->getID()));
						if (it != is_upgrading.end()) {
							onUpgradeEnd(*pl, *up, frame);
							is_upgrading.erase(it);
						}
					}
				}
			}
		}

	public:
		Trace trace; 

		void onStart() {
			trace_name = BWAPI::Broodwar->mapFileName();

			for (std::set<BWAPI::Player*>::iterator it = BWAPI::Broodwar->getPlayers().begin(); it != BWAPI::Broodwar->getPlayers().end(); ++it) {
				create_player(*it);
			}

			trace.game.map_name = BWAPI::Broodwar->mapName();
			trace.game.frames = BWAPI::Broodwar->getReplayFrameCount();
			trace.game.n_players = players.size();
		}

		void onFrame() {
			int frame = BWAPI::Broodwar->getFrameCount();
			for (std::map<BWAPI::Player*, size_t>::iterator it = players.begin(); it != players.end(); ++it) {
				player_on_frame(frame, it->first, trace.players[it->second]);
			}

			for (std::vector<BWAPI::Unit*>::iterator it = traced_units.begin(); it != traced_units.end(); ++it) {
				BWAPI::Unit* unit = *it;
				unit_on_frame(unit, frame, trace.units[units[unit]]);
			} 

			for (std::vector<BWAPI::Unit*>::iterator it = traced_buildings.begin(); it != traced_buildings.end(); ++it) {
				building_on_frame(*it, frame, trace.buildings[buildings[*it]]);
			}

			check_upgrades();
		}

		void onEnd(BWAPI::Player* player) {
			trace.winner = player->getName();
		}

		bool isUnit(BWAPI::Unit* unit) {
			return (!(unit->getPlayer()->isNeutral()) &&
					!(unit->getType().isBuilding()) &&
					!(unit->isHallucination()) &&
					!(unit->getType().isSpell()));
		}

		bool isBuilding(BWAPI::Unit* unit) {
			return (!(unit->getPlayer()->isNeutral()) &&
					 (unit->getType().isBuilding()) &&
					!(unit->getType().isSpell()));
		}

		void onUpgradeStart(BWAPI::Player* player, const BWAPI::UpgradeType& upgrade, int frame) {
			UpgradeTrace t;
			t.start = frame;
			t.type_id = upgrade.getID();
			t.type_name = upgrade.getName();
			t.owner = player->getID();
			t.level = player->getUpgradeLevel(upgrade) + 1;
			t.end = -1;

			trace.upgrades.push_back(t);
		}

		void onUpgradeEnd(BWAPI::Player* player, const BWAPI::UpgradeType& upgrade, int frame) {
			for (std::vector<UpgradeTrace>::iterator it = trace.upgrades.begin(); it != trace.upgrades.end(); ++it) {
				if ((it->owner == player->getID()) && (it->type_id == upgrade.getID()) && (it->end == -1)) {
					it->end = frame;
				}
			}
		}

		void onUnitCreate(BWAPI::Unit* unit) {
			int frame = BWAPI::Broodwar->getFrameCount();
			//If it is a larva, egg, siege tank siege mode, care
			if(unit->getType().getID() == 30 || unit->getType().getID() == 35 || unit->getType().getID() == 36) {return;}
			if (isUnit(unit)) {
				traced_units.push_back(unit);
				create_unit(unit, frame);
			}
			if (isBuilding(unit)) {
				traced_buildings.push_back(unit);
				create_building(unit, frame); 
			}
		}

		void onUnitMorph(BWAPI::Unit* unit) {
			int frame = BWAPI::Broodwar->getFrameCount();
			//Tank to tank mode or tank to siege mode is of no interest, and neither is zerg egg
			if(unit->getType().getID() == 5 || unit->getType().getID() == 30 || unit->getType().getID() == 35 || unit->getType().getID() == 36) {return;}

			if (isUnit(unit)) {
				traced_units.push_back(unit);
				create_unit(unit, frame);
			}
			if (isBuilding(unit) && !(buildings.count(unit) == 1)) {
				traced_buildings.push_back(unit);
				create_building(unit, frame);
			}
		}

		void onUnitDestroy(BWAPI::Unit* unit) {
			int frame = BWAPI::Broodwar->getFrameCount();
			if (isUnit(unit)) {
				std::vector<BWAPI::Unit*>::iterator it = std::find(traced_units.begin(), traced_units.end(), unit);
				if (it != traced_units.end()) {
					trace.units[units[unit]].death = frame;
					traced_units.erase(it);
				}
			} else if (isBuilding(unit)) {
				std::vector<BWAPI::Unit*>::iterator it = std::find(traced_buildings.begin(), traced_buildings.end(), unit);
				if (it != traced_buildings.end()) {
					trace.buildings[buildings[*it]].death = frame;
					traced_buildings.erase(it);
				}
			}
		}
	};
}