
#include <vector>
#include <string>
#include <stdlib.h>
#include <time.h>

#include "util.h"
#include "TestRunner.h"
#include "TestRunPrinter.h"
#include "Database.h"

const char* usage = "testrunner (command) [args]\n"
					" where command is one of the following:\n"
					"  gen-ini,  args: testfile\n"
					"  test-parse,  args: testfile\n"
					"  reset-test,  args: testfile\n";

std::vector<std::string> parseargs(int argc, const char* argv[]) {
	std::vector<std::string> args;
	for (int i = 1; i != argc; ++i)
		args.push_back(std::string(argv[i]));
	return args;
}

void gen_ini(const std::string& testfile, const std::string& inifile) {
	std::cout << "Generating bwapi.ini" << std::endl;
	scai::BwapiIni ini;
	ini.load(inifile);

	ini.auto_menu = "SINGLE_PLAYER";
	ini.auto_restart = "ON";
	ini.sound = "OFF";

	scai::TestFileParser p;
	p.parse(testfile);
	ini.map = p.map;
	ini.enemy_race = p.race;

	ini.save(inifile);
}

void test_parser(const std::string& testfile) {
	scai::TestFileParser p;
	p.parse(testfile);

	std::cout << "Runs: " << p.runs << '\n';
	std::cout << "Race: " << p.race << '\n';
	std::cout << "Database: " << p.database << '\n';
	std::cout << "Map: " << p.map << std::endl;
}

void reset_test(const std::string& testfile) {
	scai::TestRunner t;
	t.load_test(testfile);
	t.reset();
}

void print_test(const std::string& testfile, const std::string& level) {
	try {
		int lvl = util::parse(level);
		scai::TestRunPrinter printer(testfile);
		printer.print(lvl, std::cout);
	} catch (std::exception e) {
		std::cerr << e.what() << std::endl;
	}
}

void print_timings(const std::string& testfile) {
	try {
		scai::TestRunPrinter printer(testfile);
		printer.print_timings(std::cout);
	} catch (std::exception e) {
		std::cerr << e.what() << std::endl;
	}
}

void print_matches(const std::string& testfile) {
	try {
		scai::TestRunPrinter printer(testfile);
		printer.print_matches(std::cout);
	} catch (std::exception e) {
		std::cerr << e.what() << std::endl;
	}
}

void join(const std::string& db1, const std::string& db2) {
	std::cout << "Reading database 1";
	scai::Database d1(db1);
	std::cout << " .. DONE" << std::endl;

	std::cout << "Reading database 2";
	scai::Database d2(db2);
	std::cout << " .. DONE" << std::endl;

	std::cout << "Joining databases";
	d1.join(d2);
	std::cout << " .. DONE" << std::endl;

	std::cout << "Writing database";
	d1.write_database(db1);
	std::cout << " .. DONE" << std::endl;
}

void resize(const std::string& in_db, const std::string& out_db, size_t new_size) {
	std::srand( std::time( NULL ) );
	std::cout << "Reading db";
	scai::Database db(in_db);
	std::cout << " .. DONE" << std::endl;
	try {
		db.resize(new_size);
	} catch (std::exception e) {
		std::cerr << e.what() << std::endl;
	}
	db.write_database(out_db);
}

void db_print(const std::string& database) {
	scai::Database db(database);
	db.print();
}

int main(int argc, const char* argv[]) {
	std::vector<std::string> args = parseargs(argc, argv);

	if (args.size() == 0) {
		std::cerr << usage << std::endl;
		return 1;
	} else if (args[0] == "gen-ini") {
		std::cout << "Generating ini with testfile: " << args[1] << std::endl;
		gen_ini(args[1], args[2]);
	} else if (args[0] == "test-parse") {
		test_parser(args[1]);
	} else if (args[0] == "reset-test") {
		reset_test(args[1]);
	} else if (args[0] == "print-test") {
		print_test(args[1], args[2]);
	} else if (args[0] == "join") {
		join(args[1], args[2]);
	} else if (args[0] == "resize") {
		resize(args[1], args[2], util::parse(args[3]));
	} else if (args[0] == "print-db") {
		db_print(args[1]);
	} else if (args[0] == "print-timing") {
		print_timings(args[1]);
	} else if (args[0] == "print-matches") {
		print_matches(args[1]);
	}

	return 0;
}