#pragma once

#include <iostream>
#include <string>
#include <boost/archive/text_iarchive.hpp>
#include "Case.h"
#include "Trace.h"
#include "Constants.h"
#include "PlanItem.h"
#include "log.h"

class Analyzer {
	/* Generic events to create PlanItems from */
	enum EventType {Train, Build, Research, Attack, Scout, Expand, End};
	struct Event {
		int event_type;
		int frame;
		int type_id;
		std::string what;
	};
	static bool cmp(const Event& a, const Event& b) {
		return a.frame < b.frame;
	}
	/* End generic event */

	struct LeftBase {
		static const int THRESHOLD = 1500;
		int who;
		int when;
		int what;
	};
	int distance(const scai::UnitTrace::UnitFrame& f, const scai::PlayerTrace& p) {
		int dx = f.pos_x - 32*p.start_x;
		int dy = f.pos_y - 32*p.start_y;
		return (int) std::sqrt((double)dx * dx + dy * dy);
	}
	bool when_left_base(int& frame, size_t unit_id, const scai::PlayerTrace& p) {
		for (size_t i = 0; i != trace.units[unit_id].frames.size(); ++i) {
			if (distance(trace.units[unit_id].frames[i], p) > LeftBase::THRESHOLD) {
				frame = trace.units[unit_id].frames[i].frame;
				return true;
			}
		}
		return false;
	}
	static bool lb_cmp(const LeftBase& a, const LeftBase& b) {
		return a.when < b.when;
	}
	std::vector<LeftBase> find_left_base(int player = 0) {
		std::vector<LeftBase> lbs;
		for (size_t i = 0; i != trace.units.size(); ++i) {
			if (trace.units[i].owner == player) {
				int left;
				if (when_left_base(left, i, trace.players[player])) {
					LeftBase lb;
					lb.who = i;
					lb.when = left;
					lb.what = trace.units[i].type_id;
					lbs.push_back(lb);
				}
			}
		}
		std::sort(lbs.begin(), lbs.end(), Analyzer::lb_cmp);
		return lbs;
	}

	static std::vector<Event> merge(const std::vector<Event>& one, const std::vector<Event>& two) {
		std::vector<Event> ret;
		for (size_t i = 0; i != one.size(); ++i)
			ret.push_back(one[i]);
		for (size_t i = 0; i != two.size(); ++i)
			ret.push_back(two[i]);
		return ret;
	}

	scai::CaseLibrary cases;
	scai::Trace trace;

	bool is_worker(int type_id) {
		return ((type_id == PROBE_ID) ||
			    (type_id == SCV_ID) ||
				(type_id == DRONE_ID));
	}

	bool is_main(int unit_id) {
		return unit_id == NEXUS_ID ||
			unit_id == CMD_ID ||
			unit_id == HATCH_ID ||
			unit_id == LAIR_ID ||
			unit_id == HIVE_ID;
	}
public:
	std::vector<size_t> valid_players() {
		std::vector<size_t> players;
		for (size_t i = 0; i != trace.players.size(); ++i) {
			if (trace.players[i].race_id == 2) // 2 == Protoss
				players.push_back(i);
		}
		return players;
	}
	void resource_rate(int frame, float& min, float& gas, int p = 0) {
		int prev = std::max(frame - 1440, 0);
		size_t i = 0;
		while (i < trace.players[p].frames.size()-1 && trace.players[p].frames[i].frame < frame) i++;
		int j = i;
		while (j > 0 && trace.players[p].frames[j].frame > prev) {
			j--;
		}

		int d_min = trace.players[p].frames[i].minerals - trace.players[p].frames[j].minerals;
		int d_gas = trace.players[p].frames[i].gas - trace.players[p].frames[j].gas;
		int d_frames = trace.players[p].frames[i].frame - trace.players[p].frames[j].frame;
		try {
			min = static_cast<float>(d_min) / d_frames;
			gas = static_cast<float>(d_gas) / d_frames;
		} catch(...) {
			std::cout << "Error in resource rate, tried to divide by zero" <<std::endl;
			throw 42;
		}
	}
	scai::Gamestate create_gamestate_at(int frame, int player = 0) {
		scai::Gamestate gs;
		gs.map = trace.game.map_name;
		gs.frame = frame;

		gs.my_race_id = trace.players[player].race_id;
		gs.their_race_id = trace.players[opponent(player)].race_id;

		size_t nearest_frame = 0;
		while (nearest_frame < trace.players[player].frames.size() - 1 && trace.players[player].frames[nearest_frame].frame < frame)
			++nearest_frame;
		gs.mMinerals = trace.players[player].frames[nearest_frame].minerals;
		gs.mGas = trace.players[player].frames[nearest_frame].gas;

		resource_rate(frame, gs.mMineralsRate, gs.mGasRate, player);
		/* TODO: Calculate min/gas rate */

		for (int i = 0; i != trace.units.size(); ++i) {
			if (is_worker(trace.units[i].type_id)) {
				//std::cout << "worker stuff: " << trace.units[i].type_id << " name: "<< trace.units[i].type_name.c_str() << " birth: " << trace.units[i].birth <<" death: "<< trace.units[i].death << " frames " <<frame<<std::endl;
				if(frame > trace.units[i].birth && (frame <= trace.units[i].death || trace.units[i].death == -1)) {
					if (trace.units[i].owner == player) gs.mHarvesters++;
					if (trace.units[i].owner == opponent(player)) gs.eHarvesters++;
				}
			}
		}

		for (int i = 0; i != trace.upgrades.size(); ++i) {
			const scai::UpgradeTrace& u = trace.upgrades[i];
			if (frame > u.end) {
				if (u.owner == player)
					gs.mResearch[u.type_id] = u.level;
				if (u.owner == opponent(player))
					gs.eResearch[u.type_id] = u.level;
			}
		}

		/* type_id -> count */
		std::map<int, int> my_unit_count;
		std::map<int, int> his_unit_count;
		for (int i = 0; i != trace.units.size(); ++i) {
			scai::UnitTrace unit = trace.units[i];
			scai::UnitTrace::UnitFrame f = unit.get_at(frame);
			//std::cout << "units: " << trace.units[i].type_id << " name: "<< trace.units[i].type_name.c_str() << " birth: " << trace.units[i].birth <<" death: "<< trace.units[i].death << " frames " <<frame<<std::endl;
			//To ensure that we only put in valid units, and avoid zerg eggs and larvas
			if (frame > unit.birth && (frame <= unit.death || unit.death == -1) && !(unit.type_id == 35 || unit.type_id == 36)) {
				if (unit.owner == player) my_unit_count[unit.type_id]++;
				if (unit.owner == opponent(player)) his_unit_count[unit.type_id]++;
			}
		}
		for (std::map<int, int>::const_iterator it = my_unit_count.begin();
			it != my_unit_count.end(); ++it)
		{
			scai::Gamestate::Units units;
			units.mine = true;
			units.count = it->second;
			units.type_id = it->first;
			gs.mUnits.push_back(units);
		}
		for (std::map<int, int>::const_iterator it = his_unit_count.begin();
			it != his_unit_count.end(); ++it)
		{
			scai::Gamestate::Units units;
			units.mine = false;
			units.count = it->second;
			units.type_id = it->first;
			gs.eUnits.push_back(units);
		}

		std::map<int, int> my_building_count;
		std::map<int, int> his_building_count;
		for (int i = 0; i != trace.buildings.size(); ++i) {
			scai::BuildingTrace b = trace.buildings[i];
			
			//std::cout << "buildings: " << trace.buildings[i].type_id << " name: "<< trace.buildings[i].type_name.c_str() << " birth: " << trace.buildings[i].birth <<" death: "<< trace.buildings[i].death << " frames " <<frame<<std::endl;
			//To ensure we only use valid buildings
			if (b.birth <= frame && (frame <= b.death || b.death == -1) ) {
				if (b.owner == player) my_building_count[b.type_id]++;
				if (b.owner == opponent(player)) his_building_count[b.type_id]++;
			}
		}
		for (std::map<int, int>::const_iterator it = my_building_count.begin();
			it != my_building_count.end(); ++it)
		{
			scai::Gamestate::Units units;
			units.mine = true;
			units.count = it->second;
			units.type_id = it->first;
			gs.mBuildings.push_back(units);
		}
		for (std::map<int, int>::const_iterator it = his_building_count.begin();
			it != his_building_count.end(); ++it)
		{
			scai::Gamestate::Units units;
			units.mine = false;
			units.count = it->second;
			units.type_id = it->first;
			gs.eBuildings.push_back(units);
		}

		int m_bases = 0, e_bases = 0;
		for (int i = 0; i != trace.buildings.size(); ++i) {
			if (is_main(trace.buildings[i].type_id)) {
				//std::cout << "bases: " << trace.buildings[i].type_id << " name: "<< trace.buildings[i].type_name.c_str() << " birth: " << trace.buildings[i].birth <<" death: "<< trace.buildings[i].death << " frames " <<frame<<std::endl;
				if(trace.buildings[i].birth <= frame && (frame <= trace.buildings[i].death || trace.buildings[i].death == -1) ) {
					if (trace.buildings[i].owner == player) gs.mBases.push_back(scai::Gamestate::Bases());
					if (trace.buildings[i].owner == opponent(player)) gs.eBases.push_back(scai::Gamestate::Bases());
				}
			}
		}

		return gs;
	}

	std::vector<Event> find_unit_create(int player = 0) {
		std::vector<Event> units;
		for (size_t i = 0; i != trace.units.size(); ++i) {
			//eggs and larvas are very uninteresting
			if (trace.units[i].type_id == 35 || trace.units[i].type_id == 36) {continue;}
			
			if (trace.units[i].owner == player) {
				Event u;
				u.event_type = Train;
				u.frame = trace.units[i].birth;
				u.type_id = trace.units[i].type_id;
				u.what = trace.units[i].type_name;
				units.push_back(u);
			}
		}
		return units;
	}

	std::vector<Event> find_structure_build(int player = 0) {
		std::vector<Event> structures;
		for (size_t i = 0; i != trace.buildings.size(); ++i) {
			if (trace.buildings[i].owner == player) {
				Event s;
				s.event_type = is_main(trace.buildings[i].type_id)? Expand : Build;
				s.type_id = trace.buildings[i].type_id;
				s.what = trace.buildings[i].type_name;
				s.frame = trace.buildings[i].birth;
				structures.push_back(s);
			}
		}
		return structures;
	}

	std::vector<Event> find_research(int player = 0) {
		std::vector<Event> res;
		for (size_t i = 0; i != trace.upgrades.size(); ++i) {
			if (trace.upgrades[i].owner == player) {
				Event e;
				e.event_type = Research;
				e.type_id = trace.upgrades[i].type_id;
				e.what = trace.upgrades[i].type_name;
				res.push_back(e);
			}
		}
		return res;
	}
public:
	int time_trace() {
		std::vector<Event> units = find_unit_create();
		//std::cout << "Found " << units.size() << " units";
		return units.size();
	}

	int opponent(int me) {
		return (me == 0)? 1 : 0;
	}
	bool isWinner(std::string player_name) {
		return trace.winner == player_name;
	}


	static const int D_ATTACK = 1440;
	scai::CaseLibrary analyze(int player = 0, std::string match_id = BAD_MATCH) {
		std::cout << "Analyzing for player " << trace.players[player].player_name;
		std::cout << " starting at " << trace.players[player].start_x;
		std::cout << ", " << trace.players[player].start_y << '\n';
		std::cout << " Race = " << trace.players[player].race_id << '\n';
		std::cout << " match id = " << match_id << '\n';

		std::vector<Event> units = find_unit_create(player);
		std::vector<Event> buildings = find_structure_build(player);
		std::vector<Event> research = find_research(player);

		std::vector<Event> all = merge(units, merge(buildings, research));

		//find opponents stuff
		std::vector<Event> opponent_units = find_unit_create(opponent(player));
		std::vector<Event> opponent_buildings = find_structure_build(opponent(player));
		std::vector<Event> opponent_research = find_research(opponent(player));

		std::vector<Event> opponent_all = merge(opponent_units, merge(opponent_buildings, opponent_research));

		
		bool scouted = false;
		int last_attack = -D_ATTACK;
		std::vector<LeftBase> foo = find_left_base(player);
		for (size_t i = 0; i != foo.size(); ++i) {
			if (!scouted && is_worker(foo[i].what)) {
				scouted = true;
				Event e;
				e.event_type = Scout;
				e.frame = foo[i].when;
				all.push_back(e);
			} else if (foo[i].when > (last_attack + D_ATTACK)) {
				last_attack = foo[i].when;
				Event e;
				e.event_type = Attack;
				e.frame = foo[i].when;
				all.push_back(e);
			}
		}

		std::sort(all.begin(), all.end(), Analyzer::cmp);
		std::sort(opponent_all.begin(), opponent_all.end(), Analyzer::cmp);

		/*for (size_t i = 0; i != all.size(); ++i) {
			if (all[i].event_type == Build) std::cout << "build " << all[i].frame << " " << all[i].type_id << " " <<all[i].what << "\n";
			else if (all[i].event_type == Train) std::cout << "train " << all[i].frame << " "  << all[i].type_id << " "<< all[i].what << "\n";
			else if (all[i].event_type == Research) std::cout << "research " << all[i].frame << " "  << all[i].type_id << " "<< all[i].what << "\n";
			else if (all[i].event_type == Scout) std::cout  << i << " " << "scout\n"<< all[i].frame;
			else if (all[i].event_type == Attack) std::cout  << i << " " << "CHARGE!!!!\n"<< all[i].frame;
			else if (all[i].event_type == Expand) std::cout  << i << " " << "Expand\n"<< all[i].frame;
			else std::cout  << i << " " << "ERROR!\n";
		}

		for (size_t i = 0; i != opponent_all.size(); ++i) {
			if (all[i].event_type == Build) std::cout << "build " << all[i].frame << " " << opponent_all[i].type_id << " " <<opponent_all[i].what << "\n";
			else if (all[i].event_type == Train) std::cout << "train " << all[i].frame << " "  << opponent_all[i].type_id << " "<< opponent_all[i].what << "\n";
			else if (all[i].event_type == Research) std::cout << "research " << all[i].frame << " "  << opponent_all[i].type_id << " "<< opponent_all[i].what << "\n";
			else if (all[i].event_type == Scout) std::cout  << i << " " << "scout\n"<< all[i].frame;
			else if (all[i].event_type == Attack) std::cout  << i << " " << "CHARGE!!!!\n"<< all[i].frame;
			else if (all[i].event_type == Expand) std::cout  << i << " " << "Expand\n"<< all[i].frame;
			else std::cout  << i << " " << "ERROR!\n";
		}*/

		scai::CaseLibrary cases = split_in_cases(all, opponent_all, player);
		for (size_t i = 0; i != cases.size(); ++i) {
			if(isWinner(trace.players[player].player_name)) {
				cases[i].wins = 1;
				cases[i].matches[match_id] = 1;
			} else {
				cases[i].loss = 1;
				cases[i].matches[match_id] = -1;
			}
			
			std::stringstream thing;
			thing << "_" << i;
			cases[i].match_id = match_id + thing.str();
			if(i != 0) {
				cases[i-1].child = cases[i].match_id;
			}
			//std::cout << i << ": " << cases[i].plan.size() << '\n';
			//std::cout << cases[i].match_id << std::endl;
		}
		for (size_t i = 0; i != cases.size(); ++i) {
			//std::cout<<"Case "<< cases[i].match_id.c_str() << " has child "<<  cases[i].child.c_str() << std::endl;
		}
		
		if(isWinner(trace.players[player].player_name)) {
			std::cout << "We won!!!!"<< std::endl;
		} else if(isWinner(trace.players[opponent(player)].player_name)) {
			std::cout << "We lost :("<< std::endl;
		}else  {
			std::cout << "We couldnt find winner, should be: "<< trace.winner << std::endl;
			for(size_t i = 0 ; i < trace.players.size() ; i++) {
				std::cout << "Player at pos "<<i<<" is "<<trace.players[i].player_name<<std::endl;
			}
			//TODO: Might want to remove this for automated stuff
			throw 42;
		}

		return cases;
	}

	bool significant(const Event& e) {
		if (e.event_type == Scout) return true;
		else if (e.event_type == Attack) return true;
		else if (e.event_type == Expand) return true;
		else return false;
	}

	scai::PlanItem translate_event(const Event& e) {
		scai::PlanItem pi;
		if (e.event_type == Attack) {
			pi.myType = scai::Attack;
			pi.attack.location = scai::Main;
		} else if (e.event_type == Scout) {
			pi.myType = scai::Scout;
			pi.scout.location = scai::Main;
		} else if (e.event_type == Build) {
			pi.myType = scai::Build;
			pi.build.type_id = e.type_id;
		} else if (e.event_type == Train) {
			pi.myType = scai::Train;
			pi.train.type_id = e.type_id;
			pi.train.count = 1;
		} else if (e.event_type == Research) {
			pi.myType = scai::Upgrade;
			pi.upgrade.upgrade_id = e.type_id;
		} else if (e.event_type == Expand) {
			pi.myType = scai::Expand;
			pi.expand.location = scai::Expansion;
		} else {
			std::cout << "Exception at translate events, event type:" <<std::endl;
			std::cout << e.event_type << std::endl;
			throw 42;
		}
		return pi;
	}
	scai::Case translate_events(const std::vector<Event>& events, std::vector<Event>& opponent_events, size_t first, size_t last, size_t& first_opponent, int player = 0) {
		scai::Case c;
		c.state = create_gamestate_at(events[first].frame, player);
		
		for (size_t i = first; i != last; ++i) {
			
			c.plan.push_back(translate_event(events[i]));
		}
		
		//std::cout << "Want to write from " << events[first].frame << " to " << events[last-1].frame << std::endl;
		int index = first_opponent;
		for(size_t i = first_opponent ; i < opponent_events.size() ; i++) {
			if(opponent_events[i].frame > events[last-1].frame) {
				index = i;
				break;
			}
			else if(!(opponent_events[i].frame < events[first].frame)) {
				c.opponentmodel.push_back(translate_event(opponent_events[i]));
				index = i;
			}
		}
		first_opponent = index;
		return c;
	}

	scai::CaseLibrary split_in_cases(std::vector<Event>& events, std::vector<Event>& opponent_events, int player = 0) {
		scai::CaseLibrary cases;
		size_t last = 0;
		size_t first_opponent = 0;

		/*for(size_t i = 0 ; i < opponent_events.size() ; i++) {
			std::cout<<"opponent model consists of: "<<opponent_events[i].event_type << " "<<opponent_events[i].frame << " "<<opponent_events[i].type_id << " "<<opponent_events[i].what<<std::endl;
		}*/
		for (size_t i = 0; i != events.size(); ++i) {
			//filter out items at time 0
			if(events[i].frame == 0) {
				last = i+1;
			}
			else if (significant(events[i])) {
				cases.push_back(translate_events(events, opponent_events, last, i+1, first_opponent, player));
				last = i+1;
			}
			//Last thing we do before ending the match, force out a case
			else if(i == events.size()-1) {
				cases.push_back(translate_events(events, opponent_events, last, i+1, first_opponent, player));
				last = i+1;
			}
		}
		/*for(size_t i = 0 ; i < trace.units.size() ; i++) {
			std::cout<<trace.units[i].type_name << std::endl;
		}*/
		/*for(size_t i = 0 ; i < cases.size() ; i++){
			scai::Case mehcase = cases[i];
			std::cout<<"Starting plan"<<std::endl;
			for(size_t j = 0 ; j < mehcase.plan.size() ; j++){
				if (mehcase.plan[j].myType == scai::Build) std::cout << "build " << mehcase.plan[j].build.type_id<< "\n";
				else if (mehcase.plan[j].myType == scai::Train) std::cout << "train " <<mehcase.plan[j].train.type_id<<  "\n";
				else if (mehcase.plan[j].myType == scai::Upgrade) std::cout << "research " << mehcase.plan[j].upgrade.upgrade_id <<"\n";
				else if (mehcase.plan[j].myType == scai::Scout) std::cout << "scout\n";
				else if (mehcase.plan[j].myType == scai::Attack) std::cout << "CHARGE!!!!\n";
				else if (mehcase.plan[j].myType == scai::Expand) std::cout << "Expand\n";
			}
			std::cout<<"Starting opponent model"<<std::endl;
			for(size_t j = 0 ; j < mehcase.opponentmodel.size() ; j++){
				if (mehcase.opponentmodel[j].myType == scai::Build) std::cout << "build " << mehcase.opponentmodel[j].build.type_id<< "\n";
				else if (mehcase.opponentmodel[j].myType == scai::Train) std::cout << "train " <<mehcase.opponentmodel[j].train.type_id<< "\n";
				else if (mehcase.opponentmodel[j].myType == scai::Upgrade) std::cout << "research " << mehcase.opponentmodel[j].upgrade.upgrade_id<< "\n";
				else if (mehcase.opponentmodel[j].myType == scai::Scout) std::cout << "scout\n";
				else if (mehcase.opponentmodel[j].myType == scai::Attack) std::cout << "CHARGE!!!!\n";
				else if (mehcase.opponentmodel[j].myType == scai::Expand) std::cout << "Expand\n";
			}
		}*/
		
		

		return cases;
	}

	Analyzer() {}

	Analyzer(const scai::Trace& trace): trace(trace) {}

	Analyzer(const std::string& trace_name) {
		load_trace(trace_name);
	}

	void load_trace(const std::string& trace_name) {
		std::ifstream f(trace_name.c_str());
		if (!f) {
			std::cout <<"Could not find file at " << trace_name.c_str() <<"!"<< std::endl;
			throw 42;
		}
		boost::archive::text_iarchive ar(f);
		ar >> trace;
	}

	void append_cases(scai::CaseLibrary& lib) {
		for (size_t i = 0; i != cases.size(); ++i)
			lib.push_back(cases[i]);
	}
};